# -*- coding: utf-8 -*-

#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

################################################################################
## Form generated from reading UI file 'time_inspector.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)


class Ui_TimeInspector(object):
    def setupUi(self, TimeInspector):
        if not TimeInspector.objectName():
            TimeInspector.setObjectName("TimeInspector")
        TimeInspector.resize(800, 600)
        self.verticalLayout_2 = QVBoxLayout(TimeInspector)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.main_layout = QVBoxLayout()
        self.main_layout.setObjectName("main_layout")
        self.commands_layout = QHBoxLayout()
        self.commands_layout.setObjectName("commands_layout")
        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.commands_layout.addItem(self.horizontalSpacer)

        self.cbAnalysisSelection = QComboBox(TimeInspector)
        self.cbAnalysisSelection.addItem("")
        self.cbAnalysisSelection.addItem("")
        self.cbAnalysisSelection.addItem("")
        self.cbAnalysisSelection.setObjectName("cbAnalysisSelection")
        sizePolicy = QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.cbAnalysisSelection.sizePolicy().hasHeightForWidth()
        )
        self.cbAnalysisSelection.setSizePolicy(sizePolicy)

        self.commands_layout.addWidget(self.cbAnalysisSelection)

        self.pbPlot = QPushButton(TimeInspector)
        self.pbPlot.setObjectName("pbPlot")

        self.commands_layout.addWidget(self.pbPlot)

        self.horizontalSpacer_2 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.commands_layout.addItem(self.horizontalSpacer_2)

        self.lbSelectionActions = QLabel(TimeInspector)
        self.lbSelectionActions.setObjectName("lbSelectionActions")

        self.commands_layout.addWidget(self.lbSelectionActions)

        self.pbAreaToggleVisibility = QPushButton(TimeInspector)
        self.pbAreaToggleVisibility.setObjectName("pbAreaToggleVisibility")

        self.commands_layout.addWidget(self.pbAreaToggleVisibility)

        self.pbSelectionKeepData = QPushButton(TimeInspector)
        self.pbSelectionKeepData.setObjectName("pbSelectionKeepData")

        self.commands_layout.addWidget(self.pbSelectionKeepData)

        self.pbSelectionCropData = QPushButton(TimeInspector)
        self.pbSelectionCropData.setObjectName("pbSelectionCropData")

        self.commands_layout.addWidget(self.pbSelectionCropData)

        self.main_layout.addLayout(self.commands_layout)

        self.verticalLayout_2.addLayout(self.main_layout)

        self.retranslateUi(TimeInspector)

        QMetaObject.connectSlotsByName(TimeInspector)

    # setupUi

    def retranslateUi(self, TimeInspector):
        TimeInspector.setWindowTitle(
            QCoreApplication.translate("TimeInspector", "Time Inspector", None)
        )
        self.cbAnalysisSelection.setItemText(
            0,
            QCoreApplication.translate(
                "TimeInspector", "Localization number per minute", None
            ),
        )
        self.cbAnalysisSelection.setItemText(
            1,
            QCoreApplication.translate(
                "TimeInspector", "Localization precision per minute", None
            ),
        )
        self.cbAnalysisSelection.setItemText(
            2,
            QCoreApplication.translate(
                "TimeInspector", "Localization precision per minute (std err)", None
            ),
        )

        self.pbPlot.setText(QCoreApplication.translate("TimeInspector", "Plot", None))
        self.lbSelectionActions.setText(
            QCoreApplication.translate("TimeInspector", "Selection", None)
        )
        self.pbAreaToggleVisibility.setText(
            QCoreApplication.translate("TimeInspector", "Hide", None)
        )
        self.pbSelectionKeepData.setText(
            QCoreApplication.translate("TimeInspector", "Keep data", None)
        )
        self.pbSelectionCropData.setText(
            QCoreApplication.translate("TimeInspector", "Drop data", None)
        )

    # retranslateUi



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
from pathlib import Path
from typing import Union

import pyqtgraph
from pyqtgraph import AxisItem, ViewBox
from PySide6.QtCore import QRect, QRectF
from PySide6.QtGui import QImage, QPainter
from PySide6.QtWidgets import QApplication, QFileDialog

from pyminflux.state import State


def export_plot_interactive(item, parent=None):
    """Save the content of the current scene to a PNG image."""
    if isinstance(item, ViewBox):
        view_box = item
    elif isinstance(item, AxisItem):
        view_box = item.getViewBox()
    elif isinstance(item, pyqtgraph.PlotItem):
        view_box = item.getViewBox()
    else:
        return

    # Ask the user to pick a file name
    filename, ext = QFileDialog.getSaveFileName(
        parent,
        "Export filtered data",
        ".",
        "PNG images (*.png)",
    )

    # Did the user cancel?
    if filename == "":
        return

    # Make sure to add the extension
    if not filename.lower().endswith(".png"):
        filename += ".png"

    # Get the dpi from the State
    state = State()

    # Save the scene to file
    export_to_image(view_box, filename, dpi=state.plot_export_dpi)


def export_to_image(
    item: pyqtgraph.ViewBox, out_file_name: Union[Path, str], dpi: int = 600
):
    """Export current QGraphicsScene to file."""

    # Get the scene from the viewBox
    if isinstance(item, pyqtgraph.ViewBox):
        plot_item = item.parentItem()
        plot_widget = plot_item.getViewWidget()
    elif isinstance(item, pyqtgraph.PlotWidget):
        plot_widget = item
    else:
        return

    # Get primary screen resolution
    app = QApplication.instance()
    screen = app.primaryScreen()
    logical_dpi = screen.logicalDotsPerInch()

    # Final image size at given dpi
    dpi_scaling = dpi / logical_dpi
    img_size = plot_widget.size() * dpi_scaling

    # Create the image and painter
    image = QImage(img_size, QImage.Format_ARGB32)
    painter = QPainter()

    # Begin the QPainter operation
    painter.begin(image)

    # Set the viewport and window of the QPainter
    painter.setViewport(QRect(0, 0, img_size.width(), img_size.height()))

    # Paint the scene onto the image
    plot_widget.render(painter)

    # End the QPainter operation
    painter.end()

    # Save the image to a file
    image.save(out_file_name)

    # Inform
    print(f"Plot exported to {out_file_name}.")



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np
import pyqtgraph as pg
from pyqtgraph import PlotWidget
from PySide6.QtCore import QPoint, Signal, Slot
from PySide6.QtGui import QAction, QColor, QFont, Qt
from PySide6.QtWidgets import QDialog, QMenu

from pyminflux.state import State
from pyminflux.ui.helpers import export_plot_interactive
from pyminflux.ui.ui_time_inspector import Ui_TimeInspector


class TimeInspector(QDialog, Ui_TimeInspector):
    """
    A QDialog to perform temporal analysis and selection.
    """

    # Signal that the fluorophore IDs have been assigned
    fluorophore_ids_assigned = Signal(int, name="fluorophore_ids_assigned")
    processing_started = Signal(None, name="processing_started")
    processing_completed = Signal(None, name="processing_completed")
    dataset_time_filtered = Signal(None, name="dataset_time_filtered")

    def __init__(self, processor, parent):
        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_TimeInspector()
        self.ui.setupUi(self)

        # Initialize state
        self.state = State()

        # Keep a reference to the Processor
        self.processor = processor

        # Constants
        self.brush = pg.mkBrush(0, 0, 0, 255)
        self.pen = pg.mkPen(None)

        # Keep track of the x-axis limits and the selection
        self.x_axis = None
        self.selection_range = None

        # Selection region
        self.selection_region = None

        # Cache plot data
        self.localizations_per_unit_time_cache = None
        self.localization_precision_per_unit_time_cache_x = None
        self.localization_precision_per_unit_time_cache_y = None
        self.localization_precision_per_unit_time_cache_z = None
        self.localization_precision_stderr_per_unit_time_cache_x = None
        self.localization_precision_stderr_per_unit_time_cache_y = None
        self.localization_precision_stderr_per_unit_time_cache_z = None

        # Time resolution
        # @TODO: This should be user definable!
        self.time_resolution_sec = 60

        # Remember visibility status of the region
        self.roi_visible = True

        # Create the main plot widget
        self.plot_widget = PlotWidget(parent=self, background="w", title="")

        # Add it to the UI
        self.ui.main_layout.addWidget(self.plot_widget)

        # Add connections
        self.ui.pbPlot.clicked.connect(self.plot_selected)
        self.ui.pbAreaToggleVisibility.clicked.connect(self.toggle_region_visibility)
        self.ui.pbSelectionKeepData.clicked.connect(self.keep_time_region)
        self.ui.pbSelectionCropData.clicked.connect(self.crop_time_region)
        self.processing_started.connect(self.disable_ui_elements)
        self.processing_completed.connect(self.enable_ui_elements)

        # Plot the dcr histogram
        self.plot_selected()

    def invalidate_cache(self):
        """Invalidate the cache."""
        self.localizations_per_unit_time_cache = None
        self.localization_precision_per_unit_time_cache_x = None
        self.localization_precision_per_unit_time_cache_y = None
        self.localization_precision_per_unit_time_cache_z = None
        self.localization_precision_stderr_per_unit_time_cache_x = None
        self.localization_precision_stderr_per_unit_time_cache_y = None
        self.localization_precision_stderr_per_unit_time_cache_z = None

    @Slot(None, name="update")
    def update(self):
        """Update the plots as response to data changes."""

        # Invalidate cache
        self.invalidate_cache()

        # Switch back to the fastest plot
        self.ui.cbAnalysisSelection.setCurrentIndex(0)

        # Plot
        self.plot_selected()

    @Slot(None, name="plot_selected")
    def plot_selected(self):
        """Perform and plot the results of the selected analysis."""

        # Do we have something to plot?
        if self.processor is None or self.processor.full_dataframe is None:
            return

        # Inform that processing started
        self.processing_started.emit()

        # It one or more charts already exists, remove them
        for item in self.plot_widget.allChildItems():
            self.plot_widget.removeItem(item)

        # Get the requested plot
        if self.ui.cbAnalysisSelection.currentIndex() == 0:
            self.plot_localizations_per_unit_time()
        elif self.ui.cbAnalysisSelection.currentIndex() == 1:
            self.plot_localization_precision_per_unit_time()
        elif self.ui.cbAnalysisSelection.currentIndex() == 2:
            self.plot_localization_precision_per_unit_time(std_err=True)
        else:
            raise ValueError("Unexpected plotting request.")

        self.plot_widget.setLabel("bottom", text="time (min)")
        self.plot_widget.showAxis("bottom")
        self.plot_widget.showAxis("left")
        self.plot_widget.setMouseEnabled(x=True, y=False)
        self.plot_widget.setMenuEnabled(False)

        # Create a linear region for setting filtering thresholds. We create it
        # small enough not to be in the way.
        if self.selection_range is None:
            mn, mx = np.percentile(np.arange(self.x_axis[0], self.x_axis[-1]), [25, 75])
            self.selection_range = (mn, mx)
        mn, mx = self.selection_range[0], self.selection_range[1]
        self.selection_region = pg.LinearRegionItem(
            values=[mn, mx],
            pen={"color": "b", "width": 3, "alpha": 0.5},
        )

        # Add to plot
        self.plot_widget.addItem(self.selection_region)

        # Attach labels to the region to be able to access them from callbacks.
        self.selection_region.low_thresh_label = pg.InfLineLabel(
            self.selection_region.lines[0], "{value:.2f}", position=0.95
        )
        self._change_region_label_font(self.selection_region.low_thresh_label)
        self.selection_region.high_thresh_label = pg.InfLineLabel(
            self.selection_region.lines[1], "{value:.2f}", position=0.90
        )
        self._change_region_label_font(self.selection_region.high_thresh_label)

        # Make sure it is visible
        self.selection_region.show()
        self.roi_visible = True

        # Connect signals
        self.selection_region.sigRegionChanged.connect(self.region_pos_changed)
        self.selection_region.sigRegionChangeFinished.connect(
            self.region_pos_changed_finished
        )
        self.plot_widget.scene().sigMouseClicked.connect(
            self.histogram_raise_context_menu
        )

        # Inform that processing completed
        self.processing_completed.emit()

    def plot_localizations_per_unit_time(self):
        """Plot number of localizations per unit time."""

        # Is the data cached?
        if self.localizations_per_unit_time_cache is None:
            if len(self.processor.filtered_dataframe.index) == 0:
                # No data to plot
                return

            # Create `time_resolution_sec` bins starting at 0.0.
            bin_edges = np.arange(
                start=0.0,
                stop=self.processor.filtered_dataframe["tim"].max()
                + self.time_resolution_sec,
                step=self.time_resolution_sec,
            )
            bin_centers = bin_edges[:-1] + 0.5 * self.time_resolution_sec
            bin_width = self.time_resolution_sec

            # Calculate the histogram of localizations per unit time
            self.localizations_per_unit_time_cache, _ = np.histogram(
                self.processor.filtered_dataframe["tim"].values,
                bins=bin_edges,
                density=False,
            )

            # Cache the x range
            self.x_axis = (bin_centers - 0.5 * bin_width) / self.time_resolution_sec

        # Plot the histogram
        chart = pg.BarGraphItem(
            x=self.x_axis,
            height=self.localizations_per_unit_time_cache,
            width=0.9 * (self.x_axis[1] - self.x_axis[0]),
            brush=self.brush,
        )

        # Update the plot
        self.plot_widget.setXRange(self.x_axis[0], self.x_axis[-1])
        self.plot_widget.setYRange(0.0, self.localizations_per_unit_time_cache.max())
        self.plot_widget.setLabel("left", text="Number of localizations per min")
        self.plot_widget.addItem(chart)

    def plot_localization_precision_per_unit_time(self, std_err: bool = False):
        """Plot localization precision as a function of time.

        Parameters
        ----------

        std_err: bool
            Set to True to plot the standard error instead of the standard deviation.
        """

        # Add a legend
        if self.plot_widget.plotItem.legend is None:
            self.plot_widget.plotItem.addLegend()

        # Is the data cached?
        if (
            not std_err and self.localization_precision_per_unit_time_cache_x is None
        ) or (
            std_err and self.localization_precision_stderr_per_unit_time_cache_x is None
        ):
            # Create `time_resolution_sec` bins starting at 0.0.
            bin_edges = np.arange(
                start=0.0,
                stop=self.processor.filtered_dataframe["tim"].max()
                + self.time_resolution_sec,
                step=self.time_resolution_sec,
            )
            bin_centers = bin_edges[:-1] + 0.5 * self.time_resolution_sec
            bin_width = self.time_resolution_sec

            # Allocate space for the results
            x_pr = np.zeros(len(bin_edges) - 1)
            y_pr = np.zeros(len(bin_edges) - 1)
            z_pr = np.zeros(len(bin_edges) - 1)

            # Now process all bins
            for i in range(len(bin_edges) - 1):
                time_range = (bin_edges[i], bin_edges[i + 1])
                df = self.processor.select_by_1d_range("tim", time_range)
                if len(df.index) > 0:
                    stats = self.processor.calculate_statistics_on(df)
                    if std_err:
                        x_pr[i] = stats["sx"].mean() / np.sqrt(len(stats["sx"]))
                        y_pr[i] = stats["sy"].mean() / np.sqrt(len(stats["sy"]))
                        z_pr[i] = stats["sz"].mean() / np.sqrt(len(stats["sz"]))
                    else:
                        x_pr[i] = stats["sx"].mean()
                        y_pr[i] = stats["sy"].mean()
                        z_pr[i] = stats["sz"].mean()
                else:
                    x_pr[i] = 0.0
                    y_pr[i] = 0.0
                    z_pr[i] = 0.0

            # Cache results for future plotting
            if std_err:
                self.localization_precision_stderr_per_unit_time_cache_x = x_pr
                self.localization_precision_stderr_per_unit_time_cache_y = y_pr
                self.localization_precision_stderr_per_unit_time_cache_z = z_pr
            else:
                self.localization_precision_per_unit_time_cache_x = x_pr
                self.localization_precision_per_unit_time_cache_y = y_pr
                self.localization_precision_per_unit_time_cache_z = z_pr

            # Cache the x range
            self.x_axis = (bin_centers - 0.5 * bin_width) / self.time_resolution_sec

        # It one or more charts already exists, remove them
        for item in self.plot_widget.allChildItems():
            self.plot_widget.removeItem(item)

        # Alias
        if std_err:
            x_pr = self.localization_precision_stderr_per_unit_time_cache_x
            y_pr = self.localization_precision_stderr_per_unit_time_cache_y
            z_pr = self.localization_precision_stderr_per_unit_time_cache_z
        else:
            x_pr = self.localization_precision_per_unit_time_cache_x
            y_pr = self.localization_precision_per_unit_time_cache_y
            z_pr = self.localization_precision_per_unit_time_cache_z

        # Get the max bin number for normalization
        n_max = np.max([x_pr.max(), y_pr.max(), z_pr.max()])

        if self.processor.is_3d:
            offset = 1 / 3
            bar_width = 0.9 / 3
        else:
            offset = 1 / 2
            bar_width = 0.9 / 2

        # Create the sx bar charts
        chart = pg.BarGraphItem(
            x=self.x_axis,
            height=x_pr,
            width=bar_width,
            brush="r",
            alpha=0.5,
            name="σx",
        )
        self.plot_widget.addItem(chart)

        # Create the sy bar charts
        chart = pg.BarGraphItem(
            x=self.x_axis + offset,
            height=y_pr,
            width=bar_width,
            brush="b",
            alpha=0.5,
            name="σy",
        )
        self.plot_widget.addItem(chart)

        # Create the sz bar charts if needed
        if self.processor.is_3d:
            chart = pg.BarGraphItem(
                x=self.x_axis + 2 * offset,
                height=z_pr,
                width=bar_width,
                brush="k",
                alpha=0.5,
                name="σz",
            )
            self.plot_widget.addItem(chart)

        # Update the plot
        self.plot_widget.setXRange(self.x_axis[0], self.x_axis[-1])
        self.plot_widget.setYRange(0.0, n_max)
        self.plot_widget.setLabel("left", text="Localization precision (nm) per min")

    @staticmethod
    def _change_region_label_font(region_label):
        """Change the region label font style."""
        text_item = region_label.textItem
        text_item.setDefaultTextColor(QColor("gray"))
        font = text_item.font()
        font.setWeight(QFont.Bold)
        font.setPointSize(20)

    def region_pos_changed(self, item):
        """Called when the line region on one of the histogram plots is changing."""

        # This seems to be a bug in pyqtgraph. When moving a LinearRegionItems with
        # two InfLineLabels attached to the region boundaries (InfiniteLine), only
        # the label of the upper bound is automatically updated.
        region = item.getRegion()
        low_thresh_label = item.low_thresh_label
        if low_thresh_label.format == "{value:.2f}":
            value = f"{min(region):.2f}"
        elif low_thresh_label.format == "{value:.0f}":
            value = f"{min(region):.0f}"
        else:
            value = f"{min(region)}"
        low_thresh_label.textItem.setPlainText(value)

    def region_pos_changed_finished(self, item):
        """Called when the line region on one of the histogram plots has changed."""
        mn, mx = item.getRegion()
        self.selection_range = (mn, mx)

    def toggle_region_visibility(self, b):
        """Toggles the visibility of the line region."""

        if self.selection_region is None:
            return

        if self.roi_visible:
            self.selection_region.hide()
            self.roi_visible = False
            self.ui.pbAreaToggleVisibility.setText("Show")
        else:
            self.selection_region.show()
            self.roi_visible = True
            self.ui.pbAreaToggleVisibility.setText("Hide")

    def crop_time_region(self):
        """Filter away selected time region."""

        # Get the range
        mn, mx = self.selection_region.getRegion()

        # Make sure the selection is up-to-date
        self.selection_range = (mn, mx)

        # Convert to seconds
        mn_s = mn * self.time_resolution_sec
        mx_s = mx * self.time_resolution_sec

        # Filter
        self.processor.filter_by_1d_range_complement("tim", (mn_s, mx_s))

        # Invalidate the cache
        self.invalidate_cache()

        # Inform that the data has been filtered
        self.dataset_time_filtered.emit()

        # Update the plot
        self.plot_selected()

    def keep_time_region(self):
        """Keep the selected time region."""

        # Get the range
        mn, mx = self.selection_region.getRegion()

        # Make sure the selection is up-to-date
        self.selection_range = (mn, mx)

        # Convert to seconds
        mn_s = mn * self.time_resolution_sec
        mx_s = mx * self.time_resolution_sec

        # Filter
        self.processor.filter_by_1d_range("tim", (mn_s, mx_s))

        # Invalidate the cache
        self.invalidate_cache()

        # Inform that the data has been filtered
        self.dataset_time_filtered.emit()

        # Update the plot
        self.plot_selected()

    def disable_ui_elements(self):
        """Disable elements and inform that a processing is ongoing."""
        self.ui.cbAnalysisSelection.setEnabled(False)
        self.ui.pbPlot.setEnabled(False)
        self.ui.lbSelectionActions.setEnabled(False)
        self.ui.pbAreaToggleVisibility.setEnabled(False)
        self.ui.pbSelectionCropData.setEnabled(False)
        self.ui.pbSelectionKeepData.setEnabled(False)
        self.repaint()

    def enable_ui_elements(self):
        """Disable elements and inform that a processing is ongoing."""
        self.ui.cbAnalysisSelection.setEnabled(True)
        self.ui.pbPlot.setEnabled(True)
        self.ui.lbSelectionActions.setEnabled(True)
        self.ui.pbAreaToggleVisibility.setEnabled(True)
        self.ui.pbSelectionCropData.setEnabled(True)
        self.ui.pbSelectionKeepData.setEnabled(True)
        self.repaint()

    def histogram_raise_context_menu(self, ev):
        """Create a context menu on the plot."""
        if ev.button() == Qt.MouseButton.RightButton:
            menu = QMenu()
            export_action = QAction("Export plot")
            export_action.triggered.connect(
                lambda checked: export_plot_interactive(ev.currentItem)
            )
            menu.addAction(export_action)
            pos = ev.screenPos()
            menu.exec(QPoint(int(pos.x()), int(pos.y())))
            ev.accept()
        else:
            ev.ignore()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from PySide6.QtCore import Signal
from PySide6.QtGui import QDoubleValidator
from PySide6.QtWidgets import QDialog

from ..state import State
from .ui_roi_ranges import Ui_ROIRanges


class ROIRanges(QDialog, Ui_ROIRanges):
    # Signal that the data viewers should be updated
    data_ranges_changed = Signal(None, name="data_ranges_changed")

    def __init__(self, parent=None):
        """Constructor."""

        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_ROIRanges()
        self.ui.setupUi(self)

        # Set dialog name
        self.setWindowTitle("ROI ranges")

        # Get a reference to the state
        self.state = State()

        # Initialize the input fields
        self.update_fields()

        # Add validators
        self.ui.leCFRMin.setValidator(QDoubleValidator(decimals=2))
        self.ui.leCFRMax.setValidator(QDoubleValidator(decimals=2))
        self.ui.leEFOMin.setValidator(QDoubleValidator(decimals=0))
        self.ui.leEFOMax.setValidator(QDoubleValidator(decimals=0))

        # Set height explicitly, since only half of the widgets will be displayed
        # at any given time
        self.resize(self.width(), 95)

    def update_fields(self):
        """Force an update of the input fields."""
        self.ui.leCFRMin.setText(f"{self.state.cfr_thresholds[0]:.2f}")
        self.ui.leCFRMax.setText(f"{self.state.cfr_thresholds[1]:.2f}")
        self.ui.leEFOMin.setText(f"{self.state.efo_thresholds[0]:.0f}")
        self.ui.leEFOMax.setText(f"{self.state.efo_thresholds[1]:.0f}")

    def accept(self):
        """Override accept slot."""

        # Update ranges in the State
        cfr_min = float(self.ui.leCFRMin.text())
        cfr_max = float(self.ui.leCFRMax.text())
        if cfr_max < cfr_min:
            cfr_min, cfr_max = cfr_max, cfr_min
        efo_min = float(self.ui.leEFOMin.text())
        efo_max = float(self.ui.leEFOMax.text())
        if efo_max < efo_min:
            efo_min, efo_max = efo_max, efo_min
        self.state.cfr_thresholds = (cfr_min, cfr_max)
        self.state.efo_thresholds = (efo_min, efo_max)

        # Inform that the ranges have changed
        self.data_ranges_changed.emit()

        # Call the base class accept() method
        super().accept()

    def set_target(self, target):
        if target not in ["efo", "cfr"]:
            raise ValueError("Expected 'efo' or 'cfr' targets.")

        if target == "efo":
            self.ui.lbCFR.setVisible(False)
            self.ui.lbMinCFR.setVisible(False)
            self.ui.leCFRMin.setVisible(False)
            self.ui.lbCFRMax.setVisible(False)
            self.ui.leCFRMax.setVisible(False)
            self.ui.lbEFO.setVisible(True)
            self.ui.lbEFOMin.setVisible(True)
            self.ui.leEFOMin.setVisible(True)
            self.ui.lbEFOMax.setVisible(True)
            self.ui.leEFOMax.setVisible(True)
            self.setWindowTitle("EFO roi range")
        else:
            self.ui.lbCFR.setVisible(True)
            self.ui.lbMinCFR.setVisible(True)
            self.ui.leCFRMin.setVisible(True)
            self.ui.lbCFRMax.setVisible(True)
            self.ui.leCFRMax.setVisible(True)
            self.ui.lbEFO.setVisible(False)
            self.ui.lbEFOMin.setVisible(False)
            self.ui.leEFOMin.setVisible(False)
            self.ui.lbEFOMax.setVisible(False)
            self.ui.leEFOMax.setVisible(False)
            self.setWindowTitle("CFR roi range")



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
import warnings
from typing import Optional

import numpy as np
from numpy.fft import ifftshift
from scipy import signal
from scipy.fft import fft2
from scipy.signal import savgol_filter

from pyminflux.render import render_xy


def img_fourier_grid(dims, dtype=float):
    """This grid has center of mass at (0, 0): if used to perform convolution via fft2, it will not produce any shift!

    Reimplemented (with modifications) from:

    * [paper] Ostersehlt, L.M., Jans, D.C., Wittek, A. et al. DNA-PAINT MINFLUX nanoscopy. Nat Methods 19, 1072-1075 (2022). https://doi.org/10.1038/s41592-022-01577-1
    * [code]  https://zenodo.org/record/6563100

    Parameters
    ----------

    dims: tuple
        Size in each dimension.
        For 1D, dims = [nx]
        For 2D, dims = [nx, ny]
        For 3D, dims = [nx, ny, nz]

    dtype: np.dtype (Optional, default float = np.float64)
        Data type of the grid.

    Returns
    -------

    x: np.ndarray
        Linear grid of coordinates if ndims == 1.
        2D mesh grid of coordinates if ndims == 2.
        3D mesh grid of coordinates if ndims == 3.

    y: np.ndarray
        2D mesh grid of coordinates if ndims == 2.
        3D mesh grid of coordinates if ndims == 3.

    z: np.ndarray
        3D mesh grid of coordinates if ndims == 3.
    """

    number_dimensions = len(dims)

    if number_dimensions == 1:
        gx = np.arange(1, dims[0] + 1).astype(dtype)
        xi = ifftshift(gx)
        xi = xi - xi[0]
        return xi

    elif number_dimensions == 2:
        gx = np.arange(1, dims[0] + 1).astype(dtype)
        gy = np.arange(1, dims[1] + 1).astype(dtype)
        xi, yi = np.meshgrid(gx, gy, indexing="ij")
        xi = ifftshift(xi)
        xi = xi - xi[0, 0]
        yi = ifftshift(yi)
        yi = yi - yi[0, 0]
        return xi, yi

    elif number_dimensions == 3:
        gx = np.arange(1, dims[0] + 1).astype(dtype)
        gy = np.arange(1, dims[1] + 1).astype(dtype)
        gz = np.arange(1, dims[2] + 1).astype(dtype)
        xi, yi, zi = np.meshgrid(gx, gy, gz, indexing="ij")
        xi = ifftshift(xi)
        xi = xi - xi[0, 0, 0]
        yi = ifftshift(yi)
        yi = yi - yi[0, 0, 0]
        zi = ifftshift(zi)
        zi = zi - zi[0, 0, 0]
        return xi, yi, zi

    else:
        raise ValueError("Unsupported dimensionality!")


def img_fourier_ring_correlation(
    image1,
    image2,
    sx: float = 1.0,
    sy: float = 1.0,
    kernel: Optional[np.ndarray] = None,
):
    """Perform Fourier ring correlation analysis on two images and returns the estimated resolution in m.

    Reimplemented (with modifications) from:

    * [paper] Ostersehlt, L.M., Jans, D.C., Wittek, A. et al. DNA-PAINT MINFLUX nanoscopy. Nat Methods 19, 1072-1075 (2022). https://doi.org/10.1038/s41592-022-01577-1
    * [code]  https://zenodo.org/record/6563100

    Parameters
    ----------

    image1: np.ndarray
        First image, possibly generated by `pyminflux.render.render_xy()`.

    image2: np.ndarray
        Second image, possibly generated by `pyminflux.render.render_xy()`.

    sx: float (Default = 1.0 nm)
        Resolution in x direction (in nm) of the rendered image to be used for calculating FRC.

    sy: float (Default = 1.0 nm)
        Resolution in x direction (in nm) of the rendered image to be used for calculating FRC.

    kernel: np.ndarray (Optional)
        2D kernel for low-pass filtering the FRC. If omitted, a 31x31 Gaussian kernel with sigma = 1.0 will be used.

    Returns
    -------

    estimated_resolution: float
        Estimated image resolution in m.

    fc: np.ndarray
        Fourier Ring Correlation of `image1` and `image2`.

    qi: np.ndarray
        Array of frequencies.

    ci: np.ndarray
        Array of Fourier Ring Correlations (corresponding to the frequencies in qi)
    """

    def bin_data(qi, data):
        """Perform binning operation."""
        real = np.bincount(qi, weights=data.flatten().real)
        imag = np.bincount(qi, weights=data.flatten().imag)
        return real + 1j * imag

    if kernel is None:
        kernel = np.outer(signal.gaussian(31, std=1), signal.gaussian(31, std=1))

    # Physical size of the image (in meters!)
    physical_image_size = (image1.shape[0] * sy * 1e-9, image1.shape[1] * sx * 1e-9)

    # Calculate Fourier transforms
    f1 = fft2(image1)
    f2 = fft2(image2)

    # Calculate derived quantities for correlation
    a = f1 * np.conj(f2)
    b = f1 * np.conj(f1)
    c = f2 * np.conj(f2)

    # 2D image representation (first smooth, then real/absolute value)
    a_sm = signal.fftconvolve(ifftshift(a), kernel, mode="same")
    b_sm = signal.fftconvolve(ifftshift(b), kernel, mode="same")
    c_sm = signal.fftconvolve(ifftshift(c), kernel, mode="same")
    fc = a_sm / np.sqrt(b_sm * c_sm)
    fc = np.real(fc)

    # Calculate frequency space grid
    qx, qy = img_fourier_grid(image1.shape)
    qx = qx / physical_image_size[0]
    qy = qy / physical_image_size[1]
    q = np.sqrt(qx**2 + qy**2)

    # Calculate bin a, b, c in dependence of q and hardcoded value B = 5e5. This value seems to give
    # more stable results than when making it a function of the frequency.
    # See original MATLAB code.
    B = 5e5  # bin size (in pixel in fourier space)
    qi = np.round(q / B).astype(int)
    idx = qi.flatten()
    qi = np.arange(0, np.max(qi) + 1) * B
    aj = bin_data(idx, a)
    bj = bin_data(idx, b)
    cj = bin_data(idx, c)

    # Calculate correlation
    with warnings.catch_warnings():
        # Any attempt to prevent divide-by-zero errors sets a hard boundary to the
        # lowest value of resolution that can be estimated.
        warnings.simplefilter("ignore")
        ci = np.real(aj / np.sqrt(bj * cj))

    # Clip at 80%
    idx = qi < np.max(qi) * 0.8
    qi = qi[idx]
    ci = ci[idx]

    # Additional smoothing
    ci = savgol_filter(ci, 7, 1)

    # Determine image resolution (in m)
    q_critical = qi[np.where(ci < 1 / 7)[0][0]] if np.any(ci < 1 / 7) else qi[-1]
    estimated_resolution = 1 / q_critical

    return estimated_resolution, fc, qi, ci


def estimate_resolution_by_frc(
    x: np.ndarray,
    y: np.ndarray,
    num_reps: int = 5,
    sx: float = 1.0,
    sy: float = 1.0,
    rx: Optional[tuple] = None,
    ry: Optional[tuple] = None,
    render_type: str = "histogram",
    fwhm: Optional[float] = None,
    seed: Optional[int] = None,
    return_all: bool = False,
):
    """Estimates signal resolution by averaging the results of Fourier Ring Correlation the required number of times.

    Parameters
    ----------

    x: np.ndarray
        Array of localization x coordinates.

    y: np.ndarray
        Array of localization y coordinates.

    num_reps: int (Default = 5)
        Number of time Fourier Ring Correlation analysis is run. The returned result will be  the average of the runs.

    sx: float (Default = 1.0)
        Resolution of the render in the x direction.

    sy: float (Default = 1.0)
        Resolution of the render in the y direction.

    rx: tuple (Optional)
        (min, max) boundaries for the x coordinates. If omitted, it will default to (x.min(), x.max()).

    ry: float (Optional)
        (min, max) boundaries for the y coordinates. If omitted, it will default to (y.min(), y.max()).

    render_type: str (Default = "histogram")
        Type of render to be generated. It must be one of:
            "histogram": simple 2D histogram of localization falling into each bin of size (sx, sy).
            "fixed_gaussian": sub-pixel resolution Gaussian fit. The Gaussian full-width half maximum is required
            (see below).

    fwhm: float (Optional)
        Requested full-width half maximum (FWHM) of the Gaussian kernel. If omitted, it is set to be
        3 * np.sqrt(np.power(sx, 2) + np.power(sy, 2)).

    seed: Optional[int]
        Seed for the random number generator if comparable runs are needed.

    return_all: bool (Default = False)
        Set to True to return all measurements along with their averages.

    Returns
    -------

    resolution: float
        Estimated resolution in nm (average over `num_reps`).

    qi: np.ndarray
        Array of frequencies.

    ci: np.ndarray
        Array of Fourier Ring Correlations (corresponding to the frequencies in qi)

    resolutions: np.ndarray (num_reps, )
        Each of the estimated `n_reps` resolutions. Only returned if `return_all` is True.

    cis: np.ndarray (:, n_reps)
        Array of Fourier Ring Correlations (corresponding to the frequencies in qi) for each of the `n_reps` runs.
        Only returned if `return_all` is True.
    """

    if seed is not None:
        # Initialize the random number generator
        rng = np.random.default_rng(seed)
    else:
        rng = np.random.default_rng()

    # Make sure we are working with NumPy arrays
    x = np.array(x)
    y = np.array(y)

    # Make sure to have the same ranges rx and ry bot both images
    if rx is None or ry is None:
        rx = (x.min(), x.max())
        ry = (y.min(), y.max())

    resolutions = np.zeros(num_reps)
    cis = None
    qi = None
    for r in range(num_reps):
        # Partition the data
        ix = rng.random(size=x.shape) < 0.5
        c_ix = np.logical_not(ix)

        # Create two images from (complementary) subsets of coordinates (x, y)
        h1 = render_xy(
            x[ix], y[ix], sx=sx, sy=sy, rx=rx, ry=ry, render_type=render_type, fwhm=fwhm
        )[0]
        h2 = render_xy(
            x[c_ix],
            y[c_ix],
            sx=sx,
            sy=sy,
            rx=rx,
            ry=ry,
            render_type=render_type,
            fwhm=fwhm,
        )[0]

        # Estimate the resolution using Fourier Ring Correlation
        estimated_resolution, fc, qi, ci = img_fourier_ring_correlation(
            h1, h2, sx=sx, sy=sy
        )

        # Store the estimated resolution, qis and cis
        resolutions[r] = estimated_resolution
        if cis is None:
            cis = np.zeros((len(ci), num_reps), dtype=float)
        cis[:, r] = ci

    # Now calculate average values
    resolution = np.mean(resolutions)
    ci = np.mean(cis, axis=1)

    # Return
    if return_all:
        return resolution, qi, ci, resolutions, cis
    else:
        return resolution, qi, ci


def get_localization_boundaries(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    alpha: float = 0.01,
    min_range: float = 200.0,
):
    """Return x, y, and z localization boundaries for analysis.

    Reimplemented (with modifications) from:

    * [paper] Ostersehlt, L.M., Jans, D.C., Wittek, A. et al. DNA-PAINT MINFLUX nanoscopy. Nat Methods 19, 1072-1075 (2022). https://doi.org/10.1038/s41592-022-01577-1
    * [code]  https://zenodo.org/record/6563100


    Parameters
    ----------

    x: np.ndarray
        Array of localization x coordinates.

    y: np.ndarray
        Array of localization y coordinates.

    z: np.ndarray
        Array of localization z coordinates.

    alpha: float (default = 0.01)
        Quantile to remove outliers. Must be 0.0 <= alpha <= 0.5.

    min_range: float (default = 200.0)
        Absolute minimum range in nm.

    Returns
    -------

    rx: tuple
        (min, max) boundaries for the x coordinates.

    ry: float
        (min, max) boundaries for the y coordinates.

    rz: float
        (min, max) boundaries for the z coordinates.
    """

    if alpha < 0 or alpha >= 0.5:
        raise ValueError("alpha must be 0 < alpha < 0.5.")

    # Make sure we are working with NumPy arrays
    x = np.array(x)
    y = np.array(y)
    z = np.array(z)

    # Get boundaries at the given alpha level
    rx = np.quantile(x, (alpha, 1 - alpha))
    ry = np.quantile(y, (alpha, 1 - alpha))
    rz = np.quantile(z, (alpha, 1 - alpha))

    # Minimal boundaries in case of drift correction
    d_rx = float(np.diff(rx)[0])
    if d_rx < min_range:
        rx = rx + (min_range - d_rx) / 2 * np.array([-1, 1])

    d_ry = float(np.diff(ry)[0])
    if d_ry < min_range:
        ry = ry + (min_range - d_ry) / 2 * np.array([-1, 1])

    d_rz = float(np.diff(rz)[0])
    if min_range > d_rz > 1e-6:  # Only in the 3D case
        rz = rz + (min_range - d_rz) / 2 * np.array([-1, 1])

    return rx, ry, rz



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from pathlib import Path
from typing import Union

import h5py
import numpy as np
import pandas as pd
from scipy.io import loadmat

from pyminflux.reader import NativeArrayReader


class MinFluxReader:
    __docs__ = "Reader of MINFLUX data in ~.pmx`, `.npy` or `.mat` formats."

    __slots__ = [
        "_filename",
        "_valid",
        "_unit_scaling_factor",
        "_data_array",
        "_data_df",
        "_data_full_df",
        "_valid_entries",
        "_is_3d",
        "_is_aggregated",
        "_reps",
        "_efo_index",
        "_cfr_index",
        "_eco_index",
        "_dcr_index",
        "_loc_index",
        "_tid_index",
        "_tim_index",
        "_vld_index",
        "_z_scaling_factor",
    ]

    def __init__(
        self,
        filename: Union[Path, str],
        valid: bool = True,
        z_scaling_factor: float = 1.0,
    ):
        """Constructor.

        Parameters
        ----------

        filename: Union[Path, str]
            Full path to the `.npy` or `.mat` file to read

        valid: bool (optional, default = True)
            Whether to load only valid localizations.

        z_scaling_factor: float (optional, default = 1.0)
            Refractive index mismatch correction factor to apply to the z coordinates.
        """

        # Store the filename
        self._filename: Path = Path(filename)
        if not self._filename.is_file():
            raise IOError(f"The file {self._filename} does not seem to exist.")

        # Store the valid flag
        self._valid: bool = valid

        # The localizations are stored in meters in the Imspector files and by
        # design also in the `.pmx` format. Here, we scale them to be in nm
        self._unit_scaling_factor: float = 1e9

        # Store the z correction factor
        self._z_scaling_factor: float = z_scaling_factor

        # Initialize the data
        self._data_array = None
        self._data_df = None
        self._data_full_df = None
        self._valid_entries = None

        # Whether the acquisition is 2D or 3D
        self._is_3d: bool = False

        # Whether the file contains aggregate measurements
        self._is_aggregated: bool = False

        # Indices dependent on 2D or 3D acquisition
        self._reps: int = -1
        self._efo_index: int = -1
        self._cfr_index: int = -1
        self._dcr_index: int = -1
        self._eco_index: int = -1
        self._loc_index: int = -1

        # Constant indices
        self._tid_index: int = 0
        self._tim_index: int = 0
        self._vld_index: int = 0

        # Load the file
        self._load()

    @property
    def z_scaling_factor(self) -> float:
        """Returns the scaling factor for the z coordinates."""
        return self._z_scaling_factor

    @property
    def is_3d(self) -> bool:
        """Returns True is the acquisition is 3D, False otherwise."""
        return self._is_3d

    @property
    def is_aggregated(self) -> bool:
        """Returns True is the acquisition is aggregated, False otherwise."""
        return self._is_aggregated

    @property
    def num_valid_entries(self) -> int:
        """Number of valid entries."""
        if self._data_array is None:
            return 0
        return self._valid_entries.sum()

    @property
    def num_invalid_entries(self) -> int:
        """Number of valid entries."""
        if self._data_array is None:
            return 0
        return np.logical_not(self._valid_entries).sum()

    @property
    def valid_raw_data(self) -> Union[None, np.ndarray]:
        """Return the raw data."""
        if self._data_array is None:
            return None
        return self._data_array[self._valid_entries].copy()

    @property
    def processed_dataframe(self) -> Union[None, pd.DataFrame]:
        """Return the raw data as dataframe (some properties only)."""
        if self._data_df is not None:
            return self._data_df

        self._data_df = self._process()
        return self._data_df

    @property
    def raw_data_dataframe(self) -> Union[None, pd.DataFrame]:
        """Return the raw data as dataframe (some properties only)."""
        if self._data_full_df is not None:
            return self._data_full_df
        self._data_full_df = self._raw_data_to_full_dataframe()
        return self._data_full_df

    @property
    def filename(self) -> Union[Path, None]:
        """Return the filename if set."""
        if self._filename is None:
            return None
        return Path(self._filename)

    @classmethod
    def processed_properties(cls) -> list:
        """Returns the properties read from the file that correspond to the processed dataframe column names."""
        return [
            "tid",
            "tim",
            "x",
            "y",
            "z",
            "efo",
            "cfr",
            "eco",
            "dcr",
            "dwell",
            "fluo",
        ]

    @classmethod
    def raw_properties(cls) -> list:
        """Returns the properties read from the file and dynamic that correspond to the raw dataframe column names."""
        return ["tid", "aid", "vld", "tim", "x", "y", "z", "efo", "cfr", "eco", "dcr"]

    def _load(self) -> bool:
        """Load the file."""

        if not self._filename.is_file():
            print(f"File {self._filename} does not exist.")
            return False

        # Call the specialized _load_*() function
        if self._filename.name.lower().endswith(".npy"):
            try:
                data_array = np.load(str(self._filename))
                if "fluo" in data_array.dtype.names:
                    self._data_array = data_array
                else:
                    self._data_array = self._migrate_npy_array(data_array)
            except ValueError as e:
                print(f"Could not open {self._filename}: {e}")
                return False

        elif self._filename.name.lower().endswith(".mat"):
            try:
                self._data_array = self._convert_from_mat()
            except ValueError as e:
                print(f"Could not open {self._filename}: {e}")
                return False

        elif self._filename.name.lower().endswith(".pmx"):
            try:
                self._data_array = NativeArrayReader().read(self._filename)
                if self._data_array is None:
                    print(f"Could not open {self._filename}.")
                    return False
            except ValueError as e:
                print(f"Could not open {self._filename}: {e}")
                return False

        else:
            print(f"Unexpected file {self._filename}.")
            return False

        # Store a logical array with the valid entries
        self._valid_entries = self._data_array["vld"]

        # Cache whether the data is 2D or 3D and whether is aggregated
        num_locs = self._data_array["itr"].shape[1]
        if num_locs == 10:
            self._is_aggregated = False
            self._is_3d = True
        elif num_locs == 5:
            self._is_aggregated = False
            self._is_3d = False
        elif num_locs == 1:
            self._is_aggregated = True
            self._is_3d = np.nanmean(self._data_array["itr"]["loc"][:, :, 2]) != 0.0
        else:
            print(f"Unexpected number of localizations per trace ({num_locs}).")
            return False

        # Set all relevant indices
        self._set_all_indices()

        # Return success
        return True

    def _convert_from_mat(self) -> Union[np.ndarray, None]:
        """Load the MAT file."""

        try:
            mat_array = loadmat(str(self._filename))
        except ValueError as e:
            print(f"Could not open {self._filename}: {e}")
            return None

        # Number of entries
        n_entries = len(mat_array["itr"]["itr"][0][0])

        # Number of iterations
        n_iters = mat_array["itr"]["itr"][0][0].shape[-1]

        # Initialize an empty structure NumPy data array
        data_array = self._create_empty_data_array(n_entries, n_iters)

        # Copy the data over
        data_array["vld"] = mat_array["vld"].ravel().astype(data_array.dtype["vld"])
        data_array["sqi"] = mat_array["sqi"].ravel().astype(data_array.dtype["sqi"])
        data_array["gri"] = mat_array["gri"].ravel().astype(data_array.dtype["gri"])
        data_array["tim"] = mat_array["tim"].ravel().astype(data_array.dtype["tim"])
        data_array["tid"] = mat_array["tid"].ravel().astype(data_array.dtype["tid"])
        data_array["act"] = mat_array["act"].ravel().astype(data_array.dtype["act"])
        data_array["dos"] = mat_array["dos"].ravel().astype(data_array.dtype["dos"])
        data_array["sky"] = mat_array["sky"].ravel().astype(data_array.dtype["sky"])
        data_array["itr"]["itr"] = mat_array["itr"]["itr"][0][0].astype(
            data_array["itr"]["itr"].dtype
        )
        data_array["itr"]["tic"] = mat_array["itr"]["tic"][0][0].astype(
            data_array["itr"]["tic"].dtype
        )
        data_array["itr"]["loc"] = mat_array["itr"]["loc"][0][0].astype(
            data_array["itr"]["loc"].dtype
        )
        data_array["itr"]["lnc"] = mat_array["itr"]["lnc"][0][0].astype(
            data_array["itr"]["lnc"].dtype
        )
        data_array["itr"]["eco"] = mat_array["itr"]["eco"][0][0].astype(
            data_array["itr"]["eco"].dtype
        )
        data_array["itr"]["ecc"] = mat_array["itr"]["ecc"][0][0].astype(
            data_array["itr"]["ecc"].dtype
        )
        data_array["itr"]["efo"] = mat_array["itr"]["efo"][0][0].astype(
            data_array["itr"]["efo"].dtype
        )
        data_array["itr"]["efc"] = mat_array["itr"]["efc"][0][0].astype(
            data_array["itr"]["efc"].dtype
        )
        data_array["itr"]["sta"] = mat_array["itr"]["sta"][0][0].astype(
            data_array["itr"]["sta"].dtype
        )
        data_array["itr"]["cfr"] = mat_array["itr"]["cfr"][0][0].astype(
            data_array["itr"]["cfr"].dtype
        )
        data_array["itr"]["dcr"] = mat_array["itr"]["dcr"][0][0].astype(
            data_array["itr"]["dcr"].dtype
        )
        data_array["itr"]["ext"] = mat_array["itr"]["ext"][0][0].astype(
            data_array["itr"]["ext"].dtype
        )
        data_array["itr"]["gvy"] = mat_array["itr"]["gvy"][0][0].astype(
            data_array["itr"]["gvy"].dtype
        )
        data_array["itr"]["gvx"] = mat_array["itr"]["gvx"][0][0].astype(
            data_array["itr"]["gvx"].dtype
        )
        data_array["itr"]["eoy"] = mat_array["itr"]["eoy"][0][0].astype(
            data_array["itr"]["eoy"].dtype
        )
        data_array["itr"]["eox"] = mat_array["itr"]["eox"][0][0].astype(
            data_array["itr"]["eox"].dtype
        )
        data_array["itr"]["dmz"] = mat_array["itr"]["dmz"][0][0].astype(
            data_array["itr"]["dmz"].dtype
        )
        data_array["itr"]["lcy"] = mat_array["itr"]["lcy"][0][0].astype(
            data_array["itr"]["lcy"].dtype
        )
        data_array["itr"]["lcx"] = mat_array["itr"]["lcx"][0][0].astype(
            data_array["itr"]["lcx"].dtype
        )
        data_array["itr"]["lcz"] = mat_array["itr"]["lcz"][0][0].astype(
            data_array["itr"]["lcz"].dtype
        )
        data_array["itr"]["fbg"] = mat_array["itr"]["fbg"][0][0].astype(
            data_array["itr"]["fbg"].dtype
        )

        # Return success
        return data_array

    def _read_from_pmx(self) -> Union[np.ndarray, None]:
        """Load the PMX file."""

        # Open the file and read the data
        with h5py.File(self._filename, "r") as f:

            # Read the file_version attribute
            file_version = f.attrs["file_version"]

            if file_version != "1.0":
                return False

            # We only read the raw NumPy array
            data_array = f["raw/npy"][:]

        return data_array

    def _process(self) -> Union[None, pd.DataFrame]:
        """Returns processed dataframe for valid (or invalid) entries.

        Returns
        -------

        df: pd.DataFrame
            Processed data as DataFrame.
        """

        # Do we have a data array to work on?
        if self._data_array is None:
            return None

        if self._valid:
            indices = self._valid_entries
        else:
            indices = np.logical_not(self._valid_entries)

        # Extract the valid iterations
        itr = self._data_array["itr"][indices]

        # Extract the valid identifiers
        tid = self._data_array["tid"][indices]

        # Extract the valid time points
        tim = self._data_array["tim"][indices]

        # Extract the fluorophore IDs
        fluo = self._data_array["fluo"][indices]
        if np.all(fluo) == 0:
            fluo = np.ones(fluo.shape, dtype=fluo.dtype)

        # The following extraction pattern will change whether the
        # acquisition is normal or aggregated
        if self.is_aggregated:
            # Extract the locations
            loc = itr["loc"].squeeze() * self._unit_scaling_factor
            loc[:, 2] = loc[:, 2] * self._z_scaling_factor

            # Extract EFO
            efo = itr["efo"]

            # Extract CFR
            cfr = itr["cfr"]

            # Extract ECO
            eco = itr["eco"]

            # Extract DCR
            dcr = itr["dcr"]

            # Dwell
            dwell = np.around(eco / (efo / 1000.0), decimals=0)

        else:
            # Extract the locations
            loc = itr[:, self._loc_index]["loc"] * self._unit_scaling_factor
            loc[:, 2] = loc[:, 2] * self._z_scaling_factor

            # Extract EFO
            efo = itr[:, self._efo_index]["efo"]

            # Extract CFR
            cfr = itr[:, self._cfr_index]["cfr"]

            # Extract ECO
            eco = itr[:, self._eco_index]["eco"]

            # Extract DCR
            dcr = itr[:, self._dcr_index]["dcr"]

            # Calculate dwell
            dwell = np.around(eco / (efo / 1000.0), decimals=0)

        # Create a Pandas dataframe for the results
        df = pd.DataFrame(
            index=pd.RangeIndex(start=0, stop=len(tid)),
            columns=MinFluxReader.processed_properties(),
        )

        # Store the extracted valid hits into the dataframe
        df["tid"] = tid
        df["x"] = loc[:, 0]
        df["y"] = loc[:, 1]
        df["z"] = loc[:, 2]
        df["tim"] = tim
        df["efo"] = efo
        df["cfr"] = cfr
        df["eco"] = eco
        df["dcr"] = dcr
        df["dwell"] = dwell
        df["fluo"] = fluo

        return df

    def _raw_data_to_full_dataframe(self) -> Union[None, pd.DataFrame]:
        """Return raw data arranged into a dataframe."""
        if self._data_array is None:
            return None

        # Intialize output dataframe
        df = pd.DataFrame(columns=MinFluxReader.raw_properties())

        # Allocate space for the columns
        n_rows = len(self._data_array) * self._reps

        # Get all unique TIDs and their counts
        _, tid_counts = np.unique(self._data_array["tid"], return_counts=True)

        # Get all tids (repeated over the repetitions)
        tid = np.repeat(self._data_array["tid"], self._reps)

        # Create virtual IDs to mark the measurements of repeated tids
        # @TODO Optimize this!
        aid = np.zeros((n_rows, 1), dtype=np.int32)
        index = 0
        for c in np.nditer(tid_counts):
            tmp = np.repeat(np.arange(c), self._reps)
            n = len(tmp)
            aid[index : index + n, 0] = tmp
            index += n

        # Get all valid flags (repeated over the repetitions)
        vld = np.repeat(self._data_array["vld"], self._reps)

        # Get all timepoints (repeated over the repetitions)
        tim = np.repeat(self._data_array["tim"], self._reps)

        # Get all localizations (reshaped to drop the first dimension)
        loc = (
            self._data_array["itr"]["loc"].reshape((n_rows, 3))
            * self._unit_scaling_factor
        )
        loc[:, 2] = loc[:, 2] * self._z_scaling_factor

        # Get all efos (reshaped to drop the first dimension)
        efo = self._data_array["itr"]["efo"].reshape((n_rows, 1))

        # Get all cfrs (reshaped to drop the first dimension)
        cfr = self._data_array["itr"]["cfr"].reshape((n_rows, 1))

        # Get all ecos (reshaped to drop the first dimension)
        eco = self._data_array["itr"]["eco"].reshape((n_rows, 1))

        # Get all dcrs (reshaped to drop the first dimension)
        dcr = self._data_array["itr"]["dcr"].reshape((n_rows, 1))

        # Build the dataframe
        df["tid"] = tid.astype(np.int32)
        df["aid"] = aid.astype(np.int32)
        df["vld"] = vld
        df["tim"] = tim
        df["x"] = loc[:, 0]
        df["y"] = loc[:, 1]
        df["z"] = loc[:, 2]
        df["efo"] = efo
        df["cfr"] = cfr
        df["eco"] = eco
        df["dcr"] = dcr

        return df

    def _set_all_indices(self):
        """Set indices of properties to be read."""
        if self._data_array is None:
            return False

        if self.is_aggregated:
            self._reps = 1
            self._efo_index = -1  # Not used
            self._cfr_index = -1  # Not used
            self._dcr_index = -1  # Not used
            self._eco_index = -1  # Not used
            self._loc_index = -1  # Not used
        else:
            if self.is_3d:
                self._reps = 10
                self._efo_index = 9
                self._cfr_index = 6
                self._dcr_index = 9
                self._eco_index = 9
                self._loc_index = 9
            else:
                self._reps = 5
                self._efo_index = 4
                self._cfr_index = 4
                self._dcr_index = 4
                self._eco_index = 4
                self._loc_index = 4

    def _create_empty_data_array(self, n_entries: int, n_iters: int) -> np.ndarray:
        """Initializes a structured data array compatible with those exported from Imspector.

        Parameters
        ----------

        n_entries: int
            Number of localizations in the array.

        n_iters: int
            Number of iterations per localization.
            10 for 3D datasets, 5 for 2D datasets, 1 for aggregated measurements.

        Returns
        -------

        array: Empty array with the requested dimensionality.
        """

        if n_iters not in [1, 5, 10]:
            raise ValueError("`n_iters` must be one of 1, 5, 10.")

        return np.empty(
            n_entries,
            dtype=np.dtype(
                [
                    (
                        "itr",
                        [
                            ("itr", "<i4"),
                            ("tic", "<u8"),
                            ("loc", "<f8", (3,)),
                            ("lnc", "<f8", (3,)),
                            ("eco", "<i4"),
                            ("ecc", "<i4"),
                            ("efo", "<f8"),
                            ("efc", "<f8"),
                            ("sta", "<i4"),
                            ("cfr", "<f8"),
                            ("dcr", "<f8"),
                            ("ext", "<f8", (3,)),
                            ("gvy", "<f8"),
                            ("gvx", "<f8"),
                            ("eoy", "<f8"),
                            ("eox", "<f8"),
                            ("dmz", "<f8"),
                            ("lcy", "<f8"),
                            ("lcx", "<f8"),
                            ("lcz", "<f8"),
                            ("fbg", "<f8"),
                        ],
                        (n_iters,),
                    ),
                    ("sqi", "<u4"),
                    ("gri", "<u4"),
                    ("tim", "<f8"),
                    ("tid", "<i4"),
                    ("vld", "?"),
                    ("act", "?"),
                    ("dos", "<i4"),
                    ("sky", "<i4"),
                    ("fluo", "<i1"),
                ]
            ),
        )

    def _migrate_npy_array(self, data_array) -> np.ndarray:
        """Migrate the raw Imspector NumPy array into a pyMINFLUX raw array."""

        # Initialize the empty target array
        new_array = self._create_empty_data_array(
            len(data_array), data_array["itr"].shape[-1]
        )

        # Copy the data over
        for field_name in data_array.dtype.names:
            new_array[field_name] = data_array[field_name]

        # Return the migrated array
        return new_array

    def __repr__(self) -> str:
        """String representation of the object."""
        if self._data_array is None:
            return "No file loaded."

        str_valid = (
            "all valid"
            if len(self._data_array) == self.num_valid_entries
            else f"{self.num_valid_entries} valid and {self.num_invalid_entries} non valid"
        )

        str_acq = "3D" if self.is_3d else "2D"
        aggr_str = "aggregated" if self.is_aggregated else "normal"

        return (
            f"File: {self._filename.name}: "
            f"{str_acq} {aggr_str} acquisition with {len(self._data_array)} entries ({str_valid})."
        )

    def __str__(self) -> str:
        """Human-friendly representation of the object."""
        return self.__repr__()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Optional

import numpy as np
from numpy.fft import ifftshift
from scipy.fft import fftn, ifftn
from scipy.interpolate import interp1d
from scipy.optimize import minimize

from pyminflux.render import render_xy, render_xyz


def drift_correction_time_windows_2d(
    x: np.ndarray,
    y: np.ndarray,
    t: np.ndarray,
    sxy: float,
    rx: Optional[tuple] = None,
    ry: Optional[tuple] = None,
    T: Optional[float] = None,
    tid: Optional[np.ndarray] = None,
):
    """Estimate 2D drift correction based on auto-correlation.

    Reimplemented (with modifications) from:

    * [paper] Ostersehlt, L.M., Jans, D.C., Wittek, A. et al. DNA-PAINT MINFLUX nanoscopy. Nat Methods 19, 1072-1075 (2022). https://doi.org/10.1038/s41592-022-01577-1
    * [code]  https://zenodo.org/record/6563100

    Parameters
    ----------

    x: np.ndarray
        Array of localization x coordinates.

    y: np.ndarray
        Array of localization y coordinates.

    t: np.ndarray
        Array of localization time points.

    sxy: float (Default = 1.0)
        Resolution in nm in both the x and y direction.

    rx: tuple (Optional)
        (min, max) boundaries for the x coordinates. If omitted, it will default to (x.min(), x.max()).

    ry: float (Optional)
        (min, max) boundaries for the y coordinates. If omitted, it will default to (y.min(), y.max()).

    T: float (Optional)
        Time window for analysis.

    tid: np.ndarray (Optional)
        Only used if T is None. The unique trace IDs are used to calculate the time window for
        analysis using some heuristics.
    """

    if T is None and tid is None:
        raise ValueError("If T is not defined, the array of TIDs must be provided.")

    # Make sure we are working with NumPy arrays
    x = np.array(x)
    y = np.array(y)
    t = np.array(t)
    if tid is not None:
        tid = np.array(tid)

    # Make sure we have valid ranges
    if rx is None:
        rx = (x.min(), x.max())

    if ry is None:
        ry = (y.min(), y.max())

    # Heuristics to define the length of the time window
    if T is None:
        rt = (t[0], t[-1])
        T = len(np.unique(tid)) * np.diff(rx)[0] * np.diff(ry)[0] / 3e6
        T = np.min([T, np.diff(rt)[0] / 2, 3600])  # At least two time windows
        T = max([T, 600])  # And at least 10 minutes long

    # Number of time windows to use
    Rt = [t[0], t[-1]]
    Ns = int(np.floor((Rt[1] - Rt[0]) / T))  # total number of time windows
    assert Ns > 1, "At least two time windows are needed, please reduce T"

    # Center of mass
    CR = 10

    # Maximum number of frames in the cross-correlation
    D = Ns

    # Weight with distance
    w = np.linspace(1, 0.5, D)

    # Regularization term (roughness)
    l = 0.1

    # get dimensions
    Nx = int(np.ceil((rx[1] - rx[0]) / sxy))
    Ny = int(np.ceil((ry[1] - ry[0]) / sxy))
    c = np.round(np.array([Ny, Nx]) / 2)

    # Create all the histograms
    h = [None] * Ns
    ti = np.zeros(Ns)  # Average times of the snapshots
    for j in range(Ns):
        t0 = Rt[0] + j * T
        idx = (t >= t0) & (t < t0 + T)
        ti[j] = np.mean(t[idx])
        hj, _, _, _ = render_xy(
            x[idx],
            y[idx],
            sx=sxy,
            sy=sxy,
            rx=rx,
            ry=ry,
            render_type="fixed_gaussian",
            fwhm=3 * sxy,
        )
        h[j] = hj

    # Compute fourier transform of every histogram
    for j in range(Ns):
        h[j] = fftn(h[j])

    # Compute cross-correlations
    dx = np.zeros((Ns, Ns))
    dy = np.zeros((Ns, Ns))
    dm = np.zeros((Ns, Ns), dtype=bool)
    dx0 = np.zeros(Ns - 1)
    dy0 = np.zeros(Ns - 1)

    for i in range(Ns - 1):
        hi = np.conj(h[i])
        for j in range(i + 1, min(Ns, i + D)):  # Either to Ns or maximally D more
            hj = ifftshift(np.real(ifftn(hi * h[j])))
            yc = c[0]
            xc = c[1]

            # Centroid estimation
            gy = np.arange(yc - 2 * CR, yc + 2 * CR + 1).astype(int)
            gx = np.arange(xc - 2 * CR, xc + 2 * CR + 1).astype(int)
            gy, gx = np.meshgrid(gy, gx, indexing="ij")
            d = hj[gy, gx]
            gy = gy.flatten()
            gx = gx.flatten()
            d = d.flatten() - np.min(d)
            d = d / np.sum(d)
            for k in range(20):
                wc = np.exp(
                    -4 * np.log(2) * ((xc - gx) ** 2 + (yc - gy) ** 2) / CR**2
                )
                n = np.sum(wc * d)
                xc = np.sum(gx * d * wc) / n
                yc = np.sum(gy * d * wc) / n
            sh = np.array([-1.0, 1.0]) * (np.array([yc, xc]) - c)
            dy[i, j] = sh[0]
            dx[i, j] = sh[1]
            dm[i, j] = True
            if j == i + 1:
                dy0[i] = sh[0]
                dx0[i] = sh[1]

    a, b = np.nonzero(dm)
    dx = dx[dm]
    dy = dy[dm]
    sx0 = np.cumsum(dx0)
    sy0 = np.cumsum(dy0)

    # Minimize cost function with some kind of regularization
    options = {"disp": False, "maxiter": 1e5}

    minimizer = lambda x: lse_distance(x, a, b, dx, w, l)
    res = minimize(minimizer, sx0, options=options, method="BFGS")
    sx = np.concatenate(([0], res.x))

    minimizer = lambda x: lse_distance(x, a, b, dy, w, l)
    res = minimize(minimizer, sy0, options=options, method="BFGS")
    sy = np.concatenate(([0], res.x))

    # Reduce by mean (so shift is minimal)
    sx = sx - np.mean(sx)
    sy = sy - np.mean(sy)

    # Multiply by pixel size
    sx = sx * sxy
    sy = sy * sxy

    # Create interpolants
    fx = interp1d(ti, sx, kind="slinear", fill_value="extrapolate")
    fy = interp1d(ti, sy, kind="slinear", fill_value="extrapolate")

    # Correct drift
    dx = fx(t)
    dy = fy(t)

    # Apply to the time points used to estimate the correction
    dxt = fx(ti)
    dyt = fy(ti)

    return dx, dy, dxt, dyt, ti, T


def drift_correction_time_windows_3d(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    t: np.ndarray,
    sxyz: float,
    rx: Optional[tuple] = None,
    ry: Optional[tuple] = None,
    rz: Optional[tuple] = None,
    T: Optional[float] = None,
    tid: Optional[np.ndarray] = None,
):
    """Estimate 3D drift correction based on auto-correlation.

    Reimplemented (with modifications) from:

    * [paper] Ostersehlt, L.M., Jans, D.C., Wittek, A. et al. DNA-PAINT MINFLUX nanoscopy. Nat Methods 19, 1072-1075 (2022). https://doi.org/10.1038/s41592-022-01577-1
    * [code]  https://zenodo.org/record/6563100

    Parameters
    ----------

    x: np.ndarray
        Array of localization x coordinates.

    y: np.ndarray
        Array of localization y coordinates.

    z: np.ndarray
        Array of localization z coordinates.

    t: np.ndarray
        Array of localization time points.

    sxyz: float (Default = 1.0)
        Resolution in nm in x, y and z direction.

    rx: tuple (Optional)
        (min, max) boundaries for the x coordinates. If omitted, it will default to (x.min(), x.max()).

    ry: float (Optional)
        (min, max) boundaries for the y coordinates. If omitted, it will default to (y.min(), y.max()).

    rz: float (Optional)
        (min, max) boundaries for the z coordinates. If omitted, it will default to (z.min(), z.max()).

    T: float (Optional)
        Time window for analysis.

    tid: np.ndarray (Optional)
        Only used if T is None. The unique trace IDs are used to calculate the time window for
        analysis using some heuristics.
    """

    if T is None and tid is None:
        raise ValueError("If T is not defined, the array of TIDs must be provided.")

    # Make sure we are working with NumPy arrays
    x = np.array(x)
    y = np.array(y)
    z = np.array(z)
    t = np.array(t)
    if tid is not None:
        tid = np.array(tid)

    # Make sure we have valid ranges
    if rx is None:
        rx = (x.min(), x.max())

    if ry is None:
        ry = (y.min(), y.max())

    if rz is None:
        rz = (z.min(), z.max())

    # Heuristics to define the length of the time window
    if T is None:
        rt = (t[0], t[-1])
        T = len(np.unique(tid)) * np.diff(rx)[0] * np.diff(ry)[0] / 3e6
        T = np.min([T, np.diff(rt)[0] / 2, 3600])  # At least two time windows
        T = max([T, 600])  # And at least 10 minutes long

    # Number of time windows to use
    Rt = [t[0], t[-1]]
    Ns = int(np.floor((Rt[1] - Rt[0]) / T))  # total number of time windows
    assert Ns > 1, "At least two time windows are needed, please reduce T"

    # Center of mass
    CR = 8

    # Maximum number of frames in the cross-correlation
    D = Ns

    # Weight with distance
    w = np.linspace(1, 0.2, D)

    # Regularization term (roughness)
    l = 0.01

    # get dimensions
    Nx = int(np.ceil((rx[1] - rx[0]) / sxyz))
    Ny = int(np.ceil((ry[1] - ry[0]) / sxyz))
    Nz = int(np.ceil((rz[1] - rz[0]) / sxyz))
    c = np.round(np.array([Nz, Ny, Nx]) / 2)

    # Create all the histograms
    h = [None] * Ns
    ti = np.zeros(Ns)  # Average times of the snapshots
    for j in range(Ns):
        t0 = Rt[0] + j * T
        idx = (t >= t0) & (t < t0 + T)
        ti[j] = np.mean(t[idx])
        hj, _, _, _, _ = render_xyz(
            x[idx],
            y[idx],
            z[idx],
            sx=sxyz,
            sy=sxyz,
            sz=sxyz,
            rx=rx,
            ry=ry,
            rz=rz,
            render_type="fixed_gaussian",
            fwhm=3 * sxyz,
        )
        h[j] = hj

    # Compute fourier transform of every histogram
    for j in range(Ns):
        h[j] = fftn(h[j])

    # Compute cross-correlations
    dx = np.zeros((Ns, Ns))
    dy = np.zeros((Ns, Ns))
    dz = np.zeros((Ns, Ns))
    dm = np.zeros((Ns, Ns), dtype=bool)
    dx0 = np.zeros(Ns - 1)
    dy0 = np.zeros(Ns - 1)
    dz0 = np.zeros(Ns - 1)

    for i in range(Ns - 1):
        hi = np.conj(h[i])
        for j in range(i + 1, min(Ns, i + D)):  # either to Ns or maximally D more
            hj = ifftshift(np.real(ifftn(hi * h[j])))
            zc = c[0]
            yc = c[1]
            xc = c[2]

            # Centroid estimation
            gz = np.arange(zc - 2 * CR, zc + 2 * CR + 1).astype(int)
            gy = np.arange(yc - 2 * CR, yc + 2 * CR + 1).astype(int)
            gx = np.arange(xc - 2 * CR, xc + 2 * CR + 1).astype(int)
            gz, gy, gx = np.meshgrid(gz, gy, gx, indexing="ij")
            d = hj[gz, gy, gx]
            gz = gz.flatten()
            gy = gy.flatten()
            gx = gx.flatten()
            d = d.flatten() - np.min(d)
            d = d / np.sum(d)
            for k in range(20):
                wc = np.exp(
                    -4
                    * np.log(2)
                    * ((xc - gx) ** 2 + (yc - gy) ** 2 + (zc - gz) ** 2)
                    / CR**2
                )
                n = np.sum(wc * d)
                xc = np.sum(gx * d * wc) / n
                yc = np.sum(gy * d * wc) / n
                zc = np.sum(gz * d * wc) / n
            sh = np.array([1.0, -1.0, 1.0]) * (np.array([zc, yc, xc]) - c)
            dz[i, j] = sh[0]
            dy[i, j] = sh[1]
            dx[i, j] = sh[2]
            dm[i, j] = True
            if j == i + 1:
                dz0[i] = sh[0]
                dy0[i] = sh[1]
                dx0[i] = sh[2]

    a, b = np.nonzero(dm)
    dx = dx[dm]
    dy = dy[dm]
    dz = dz[dm]
    sx0 = np.cumsum(dx0)
    sy0 = np.cumsum(dy0)
    sz0 = np.cumsum(dz0)

    # Minimize cost function with some kind of regularization
    options = {"disp": False, "maxiter": 1e5}

    minimizer = lambda x: lse_distance(x, a, b, dx, w, l)
    res = minimize(minimizer, sx0, options=options, method="BFGS")
    sx = np.concatenate(([0], res.x))

    minimizer = lambda x: lse_distance(x, a, b, dy, w, l)
    res = minimize(minimizer, sy0, options=options, method="BFGS")
    sy = np.concatenate(([0], res.x))

    minimizer = lambda x: lse_distance(x, a, b, dz, w, l)
    res = minimize(minimizer, sz0, options=options, method="BFGS")
    sz = np.concatenate(([0], res.x))

    # Reduce by mean (so shift is minimal)
    sx = sx - np.mean(sx)
    sy = sy - np.mean(sy)
    sz = sz - np.mean(sz)

    # Multiply by voxel size
    sx = sx * sxyz
    sy = sy * sxyz
    sz = sz * sxyz

    # Create interpolants
    fx = interp1d(ti, sx, kind="slinear", fill_value="extrapolate")
    fy = interp1d(ti, sy, kind="slinear", fill_value="extrapolate")
    fz = interp1d(ti, sz, kind="slinear", fill_value="extrapolate")

    # Correct drift
    dx = fx(t)
    dy = fy(t)
    dz = fz(t)

    # Apply to the time points used to estimate the correction
    dxt = fx(ti)
    dyt = fy(ti)
    dzt = fz(ti)

    return dx, dy, dz, dxt, dyt, dzt, ti, T


def lse_distance(x, a, b, s, w, l):
    """Calculate Least Squares Error (LSE) with regularization.

    Reimplemented (with modifications) from:

    * [paper] Ostersehlt, L.M., Jans, D.C., Wittek, A. et al. DNA-PAINT MINFLUX nanoscopy. Nat Methods 19, 1072-1075 (2022). https://doi.org/10.1038/s41592-022-01577-1
    * [code]  https://zenodo.org/record/6563100

    Parameters
    ----------

    x: np.ndarray

    a: np.ndarray
        Array of indices to be used (with b) to calculate the predicted shifts (x[b] - x[a]).

    b: np.ndarray
        Array of indices to be used (with a) to calculate the predicted shifts (x[b] - x[a]).

    s: np.ndarray
        Array of shifts to be compared.

    w: float
        Weighting factor.

    l: float
        Regularization term.
    """

    # Add shift 0 for the first frame
    x = np.concatenate(([0], x))
    dx = x[b] - x[a]
    y = w[b - a] * (dx - s) ** 2
    y = np.mean(y)

    # Add regularization term (roughness)
    r = l * np.sum(np.diff(x) ** 2)
    y = y + r

    return y



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from pathlib import Path
from typing import Union

import numpy as np

from pyminflux.processor import MinFluxProcessor


class MinFluxWriter:
    __docs__ = "Writer of (filtered) MINFLUX data to either `.npy` or `.csv` formats."

    @staticmethod
    def write_npy(processor: MinFluxProcessor, file_name: Union[Path, str]) -> bool:
        """Write (filtered) data as a NumPy structured array with the same structure as in the Imspector `.npy` file."""

        # Get the raw data
        filtered_array = processor.filtered_numpy_array
        if filtered_array is None:
            return False

        # Save the filtered structured NumPy array to disk
        try:
            np.save(str(file_name), filtered_array)
        except Exception as e:
            print(f"Could not save {file_name}: {e}")
            return False

        return True

    @staticmethod
    def write_csv(processor: MinFluxProcessor, file_name: Union[Path, str]) -> bool:
        """Write (filtered) data as a comma-separated-value `.csv` file."""

        # Save the filtered dataframe to disk
        try:
            processor.filtered_dataframe.to_csv(file_name, index=False)
        except Exception as e:
            print(f"Could not save {file_name}: {e}")
            return False

        return True



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from pathlib import Path
from typing import Union

import h5py
import pandas as pd


@dataclass
class Metadata:
    z_scaling_factor: float
    min_trace_length: int
    efo_thresholds: tuple[int, int]
    cfr_thresholds: tuple[float, float]
    num_fluorophores: int


class NativeMetadataReader:
    """Reads metadata information from `.pmx` files."""

    @staticmethod
    def scan(filename: Union[Path, str]):
        """Constructor.

        Parameters
        ----------

        filename: Union[Path, str]
            Full path to the `.pmx` file to scan.
        """

        # Open the file
        with h5py.File(filename, "r") as f:

            # Read the file_version attribute
            file_version = f.attrs["file_version"]

            if file_version != "1.0":
                return None

            # Read the parameters
            z_scaling_factor = f["parameters/z_scaling_factor"][()]
            min_trace_length = f["parameters/min_trace_length"][()]
            try:
                efo_thresholds = tuple(f["parameters/applied_efo_thresholds"][:])
            except KeyError as e:
                efo_thresholds = None
            try:
                cfr_thresholds = tuple(f["parameters/applied_cfr_thresholds"][:])
            except KeyError as e:
                cfr_thresholds = None
            num_fluorophores = f["parameters/num_fluorophores"][()]

        # Store and return
        metadata = Metadata(
            z_scaling_factor=z_scaling_factor,
            min_trace_length=min_trace_length,
            efo_thresholds=efo_thresholds,
            cfr_thresholds=cfr_thresholds,
            num_fluorophores=num_fluorophores,
        )

        return metadata


class NativeArrayReader:
    """Reads the native NumPy array from `.pmx` files."""

    @staticmethod
    def read(filename: Union[Path, str]):
        """Constructor.

        Parameters
        ----------

        filename: Union[Path, str]
            Full path to the `.pmx` file to scan.
        """

        # Open the file and read the data
        with h5py.File(filename, "r") as f:

            # Read the file_version attribute
            file_version = f.attrs["file_version"]

            if file_version != "1.0":
                return None

            # We only read the raw NumPy array
            data_array = f["raw/npy"][:]

        return data_array


class NativeDataFrameReader:
    """Reads the Pandas DataFrame from `.pmx` files."""

    @staticmethod
    def read(filename: Union[Path, str]):
        """Constructor.

        Parameters
        ----------

        filename: Union[Path, str]
            Full path to the `.pmx` file to scan.
        """

        with h5py.File(filename, "r") as f:

            # Read the file_version attribute
            file_version = f.attrs["file_version"]

            if file_version != "1.0":
                return None

            # Read dataset
            dataset = f["/paraview/dataframe"]

            # Read the NumPy data
            data_array = dataset[:]

            # Read column names
            column_names = dataset.attrs["column_names"]

            # Read column data types
            column_types = dataset.attrs["column_types"]

            # Read the index
            index_data = f["/paraview/dataframe_index"][:]

            # Create DataFrame with specified columns
            df = pd.DataFrame(data_array, index=index_data, columns=column_names)

            # Apply column data types
            for col, dtype in zip(column_names, column_types):
                df[col] = df[col].astype(dtype)

        return df



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from typing import Optional, Tuple

import numpy as np
import pyqtgraph as pg
from pyqtgraph import AxisItem, ViewBox
from PySide6 import QtCore
from PySide6.QtCore import QPoint, QSignalBlocker, Signal, Slot
from PySide6.QtGui import QAction, QColor, QDoubleValidator, QFont, Qt
from PySide6.QtWidgets import QDialog, QFileDialog, QLabel, QMenu

from ..analysis import find_cutoff_near_value, get_robust_threshold, prepare_histogram
from ..processor import MinFluxProcessor
from ..state import State
from ..utils import intersect_2d_ranges
from .helpers import export_plot_interactive
from .roi_ranges import ROIRanges
from .ui_analyzer import Ui_Analyzer


class Analyzer(QDialog, Ui_Analyzer):
    # Signal that the data viewers should be updated
    data_filters_changed = Signal(name="data_filters_changed")
    plotting_started = Signal(name="plotting_started")
    plotting_completed = Signal(name="plotting_completed")
    efo_bounds_changed = Signal(name="efo_bounds_changed")
    cfr_bounds_changed = Signal(name="cfr_bounds_changed")
    cfr_threshold_factor_changed = Signal(name="cfr_threshold_factor_changed")
    cfr_lower_bound_state_changed = Signal(name="cfr_lower_bound_state_changed")
    cfr_upper_bound_state_changed = Signal(name="cfr_upper_bound_state_changed")

    def __init__(self, minfluxprocessor: MinFluxProcessor, parent=None):
        """Constructor."""

        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_Analyzer()
        self.ui.setupUi(self)

        # Store the reference to the reader
        self.processor = minfluxprocessor

        # Reference to the communications label
        self.communication_label = None

        # Keep references to the plots
        self.efo_plot = None
        self.cfr_plot = None
        self.sx_plot = None
        self.sy_plot = None
        self.sz_plot = None
        self.efo_region = None
        self.cfr_region = None

        # Keep a reference to the singleton State class
        self.state = State()

        # Remember the data ranges (default to the options)
        self.efo_range = self.state.efo_range
        self.cfr_range = self.state.cfr_range
        self.loc_precision_range = self.state.loc_precision_range

        # Create widgets
        self.create_widgets()

        #
        # Set defaults
        #

        # EFO cutoff frequency tab
        self.ui.leEFOExpectedCutoff.setText(str(self.state.efo_expected_frequency))
        self.ui.leEFOExpectedCutoff.setValidator(
            QDoubleValidator(bottom=0.0, decimals=2)
        )

        # CFR filtering tab
        self.ui.checkCFRLowerThreshold.setChecked(self.state.enable_cfr_lower_threshold)
        self.ui.checkCFRUpperThreshold.setChecked(self.state.enable_cfr_upper_threshold)
        self.ui.leCFRFilterThreshFactor.setText(str(self.state.cfr_threshold_factor))
        self.ui.leCFRFilterThreshFactor.setValidator(
            QDoubleValidator(bottom=0.0, top=5.0, decimals=2)
        )

        # Keep a reference to the ROIRanges dialog
        self.roi_ranges_dialog = None

        # Set signal-slot connections
        self.setup_conn()

    def setup_conn(self):
        """Set up signal-slot connections."""

        # EFO peak detection tab
        self.ui.pbDetectCutoffFrequency.clicked.connect(
            self.run_efo_cutoff_frequency_detection
        )
        self.ui.pbEFORunFilter.clicked.connect(
            self.run_efo_filter_and_broadcast_viewers_update
        )
        self.ui.leEFOExpectedCutoff.textChanged.connect(
            self.persist_efo_expected_frequency
        )

        # CFR filtering tab
        self.ui.pbCFRRunFilter.clicked.connect(
            self.run_cfr_filter_and_broadcast_viewers_update
        )
        self.ui.checkCFRLowerThreshold.stateChanged.connect(
            self.persist_cfr_lower_threshold
        )
        self.ui.checkCFRUpperThreshold.stateChanged.connect(
            self.persist_cfr_upper_threshold
        )
        self.ui.pbCFRRunAutoThreshold.clicked.connect(self.run_cfr_auto_threshold)
        self.ui.leCFRFilterThreshFactor.textChanged.connect(
            self.persist_cfr_threshold_factor
        )

        # Others
        self.plotting_started.connect(self.disable_buttons)
        self.plotting_completed.connect(self.enable_buttons)

    def closeEvent(self, ev):
        """Close event."""
        if self.roi_ranges_dialog is not None:
            self.roi_ranges_dialog.close()
        super().closeEvent(ev)

    def create_widgets(self):
        """Create widgets."""

        # Communications layout
        self.communication_label = QLabel("Sorry, no data.")
        self.communication_label.setAlignment(QtCore.Qt.AlignCenter)
        font = self.communication_label.font()
        font.setPointSize(16)
        self.communication_label.setFont(font)
        self.ui.communication_layout.addWidget(self.communication_label)
        self.communication_label.hide()

        # Parameters Layout
        self.efo_plot = pg.PlotWidget(parent=self, background="w", title="EFO")
        self.efo_plot.setMouseEnabled(x=True, y=False)
        self.efo_plot.setMenuEnabled(False)
        self.efo_plot.scene().sigMouseClicked.connect(
            self.histogram_raise_context_menu_with_filtering
        )

        self.cfr_plot = pg.PlotWidget(parent=self, background="w", title="CFR")
        self.cfr_plot.setMouseEnabled(x=True, y=False)
        self.cfr_plot.setMenuEnabled(False)
        self.cfr_plot.scene().sigMouseClicked.connect(
            self.histogram_raise_context_menu_with_filtering
        )

        self.ui.parameters_layout.addWidget(self.efo_plot)
        self.efo_plot.hide()
        self.ui.parameters_layout.addWidget(self.cfr_plot)
        self.cfr_plot.hide()

        # Localizations layout
        self.sx_plot = pg.PlotWidget(parent=self, background="w", title="σx")
        self.sx_plot.setMouseEnabled(x=True, y=False)
        self.sx_plot.setMenuEnabled(False)
        self.sx_plot.scene().sigMouseClicked.connect(self.histogram_raise_context_menu)

        self.sy_plot = pg.PlotWidget(parent=self, background="w", title="σy")
        self.sy_plot.setMouseEnabled(x=True, y=False)
        self.sy_plot.setMenuEnabled(False)
        self.sy_plot.scene().sigMouseClicked.connect(self.histogram_raise_context_menu)

        self.sz_plot = pg.PlotWidget(parent=self, background="w", title="σz")
        self.sz_plot.setMouseEnabled(x=True, y=False)
        self.sz_plot.setMenuEnabled(False)
        self.sz_plot.scene().sigMouseClicked.connect(self.histogram_raise_context_menu)

        self.ui.localizations_layout.addWidget(self.sx_plot)
        self.sx_plot.hide()
        self.ui.localizations_layout.addWidget(self.sy_plot)
        self.sy_plot.hide()
        self.ui.localizations_layout.addWidget(self.sz_plot)
        self.sz_plot.hide()

    def keyPressEvent(self, ev):
        """Key press event."""

        # Do not capture key events for now
        ev.ignore()

    def set_processor(self, minfluxprocessor: MinFluxProcessor):
        """Reset the internal state."""
        self.processor = minfluxprocessor
        self.efo_region = None
        self.cfr_region = None
        self.efo_range = None
        self.cfr_range = None

    @Slot(name="run_efo_cutoff_frequency_detection")
    def run_efo_cutoff_frequency_detection(self):
        """Run EFO cutoff frequency detection."""

        # Get the approximate frequency
        frequency = 2.0 * float(self.ui.leEFOExpectedCutoff.text())

        # Histogram bin settings
        efo_auto_bins = self.state.efo_bin_size_hz == 0

        # Calculate thresholds for EFO
        n_efo, _, b_efo, _ = prepare_histogram(
            self.processor.filtered_dataframe["efo"].values,
            auto_bins=efo_auto_bins,
            bin_size=self.state.efo_bin_size_hz,
        )
        cutoff_frequency = find_cutoff_near_value(
            counts=n_efo,
            bins=b_efo,
            expected_value=frequency,
        )

        # If the search failed, we return
        if cutoff_frequency is None:
            return

        # Set the new upper bound and reset the lower bound
        min_efo = self.processor.filtered_dataframe["efo"].values.min()
        max_efo = cutoff_frequency
        self.state.efo_thresholds = (min_efo, max_efo)

        # Update plot
        self.efo_region.setRegion(self.state.efo_thresholds)

    @Slot(name="run_cfr_auto_threshold")
    def run_cfr_auto_threshold(self):
        """Run auto-threshold on EFO and CFR values and update the plots."""

        # Is there something to calculate?
        if (
            not self.state.enable_cfr_lower_threshold
            and not self.state.enable_cfr_upper_threshold
        ):
            print("Both lower and upper CFR thresholds are disabled.")
            return

        # Initialize values
        if self.state.cfr_thresholds is None:
            min_cfr = self.processor.filtered_dataframe["cfr"].values.min()
            max_cfr = self.processor.filtered_dataframe["cfr"].values.max()
        else:
            min_cfr = self.state.cfr_thresholds[0]
            max_cfr = self.state.cfr_thresholds[1]

        # Calculate thresholds for CFR
        upper_thresh_cfr, lower_thresh_cfr, _, _ = get_robust_threshold(
            self.processor.filtered_dataframe["cfr"].values,
            factor=self.state.cfr_threshold_factor,
        )
        if self.state.enable_cfr_lower_threshold:
            min_cfr = lower_thresh_cfr
        if self.state.enable_cfr_upper_threshold:
            max_cfr = upper_thresh_cfr
        self.state.cfr_thresholds = (min_cfr, max_cfr)

        # Update plot
        self.cfr_region.setRegion(self.state.cfr_thresholds)

    @Slot(int, name="persist_cfr_lower_threshold")
    def persist_cfr_lower_threshold(self, state):
        self.state.enable_cfr_lower_threshold = state != 0
        self.cfr_lower_bound_state_changed.emit()

    @Slot(int, name="persist_cfr_upper_threshold")
    def persist_cfr_upper_threshold(self, state):
        self.state.enable_cfr_upper_threshold = state != 0
        self.cfr_upper_bound_state_changed.emit()

    @Slot(str, name="persist_efo_expected_frequency")
    def persist_efo_expected_frequency(self, text):
        try:
            efo_expected_frequency = float(text)
        except ValueError as _:
            return
        self.state.efo_expected_frequency = efo_expected_frequency

    @Slot(str, name="persist_cfr_threshold_factor")
    def persist_cfr_threshold_factor(self, text):
        try:
            cfr_threshold_factor = float(text)
        except ValueError as _:
            return
        self.state.cfr_threshold_factor = cfr_threshold_factor

        # Broadcast
        self.cfr_threshold_factor_changed.emit()

    @Slot(name="run_efo_filter_and_broadcast_viewers_update")
    def run_efo_filter_and_broadcast_viewers_update(self):
        """Apply the EFO filter and inform the rest of the application that the data viewers should be updated."""

        # Apply the EFO filter if needed
        if self.state.efo_thresholds is not None:
            self.processor.filter_by_1d_range(
                "efo", (self.state.efo_thresholds[0], self.state.efo_thresholds[1])
            )

        # Update State.applied_efo_thresholds
        if self.state.applied_efo_thresholds is None:
            self.state.applied_efo_thresholds = self.state.efo_thresholds
        else:
            self.state.applied_efo_thresholds = intersect_2d_ranges(
                self.state.efo_thresholds, self.state.applied_efo_thresholds
            )

        # Update the histograms
        self.plot()

        # Signal that the external viewers should be updated
        self.data_filters_changed.emit()

    @Slot(name="run_cfr_filter_and_broadcast_viewers_update")
    def run_cfr_filter_and_broadcast_viewers_update(self):
        """Apply the CFR filter and inform the rest of the application that the data viewers should be updated."""

        # Apply the CFR filter if needed
        if self.state.cfr_thresholds is not None:
            self.processor.filter_by_1d_range(
                "cfr", (self.state.cfr_thresholds[0], self.state.cfr_thresholds[1])
            )

        # Update State.applied_cfr_thresholds
        if self.state.applied_cfr_thresholds is None:
            self.state.applied_cfr_thresholds = self.state.cfr_thresholds
        else:
            self.state.applied_cfr_thresholds = intersect_2d_ranges(
                self.state.cfr_thresholds, self.state.applied_cfr_thresholds
            )

        # Update the histograms
        self.plot()

        # Signal that the external viewers should be updated
        self.data_filters_changed.emit()

    @Slot(name="disable_buttons")
    def disable_buttons(self):
        self.ui.pbDetectCutoffFrequency.setEnabled(False)
        self.ui.pbCFRRunAutoThreshold.setEnabled(False)
        self.ui.pbEFORunFilter.setEnabled(False)
        self.ui.pbCFRRunFilter.setEnabled(False)

    @Slot(name="enable_buttons")
    def enable_buttons(self):
        self.ui.pbDetectCutoffFrequency.setEnabled(True)
        self.ui.pbCFRRunAutoThreshold.setEnabled(True)
        self.ui.pbEFORunFilter.setEnabled(True)
        self.ui.pbCFRRunFilter.setEnabled(True)

    def plot(self):
        """Plot histograms."""

        # Hide the communications label
        self.communication_label.hide()

        # Make sure there is data to plot
        is_data = True
        if self.processor is None:
            is_data = False

        if self.processor.filtered_dataframe is None:
            is_data = False

        if self.processor.num_values == 0:
            is_data = False

        # Announce that the plotting has started
        self.plotting_started.emit()

        for item in self.efo_plot.allChildItems():
            self.efo_plot.removeItem(item)
        for item in self.cfr_plot.allChildItems():
            self.cfr_plot.removeItem(item)
        for item in self.sx_plot.allChildItems():
            self.sx_plot.removeItem(item)
        for item in self.sy_plot.allChildItems():
            self.sy_plot.removeItem(item)
        for item in self.sz_plot.allChildItems():
            self.sz_plot.removeItem(item)

        # Is there data to plot?
        if not is_data:
            # Show the communication label
            self.communication_label.show()

            # Announce that the plotting has completed
            self.plotting_completed.emit()

            return

        # Histogram bin settings
        efo_auto_bins = self.state.efo_bin_size_hz == 0

        # Calculate the proper aspect ratio and view ranges for the relevant plots
        n_efo, efo_bin_edges, efo_bin_centers, efo_bin_width = prepare_histogram(
            self.processor.filtered_dataframe["efo"].values,
            auto_bins=efo_auto_bins,
            bin_size=self.state.efo_bin_size_hz,
        )
        n_cfr, cfr_bin_edges, cfr_bin_centers, cfr_bin_width = prepare_histogram(
            self.processor.filtered_dataframe["cfr"].values,
            auto_bins=True,
            bin_size=0.0,
        )

        #
        # Get "efo" and "cfr" measurements, and "sx", "sy" and "sz" localization jitter
        #

        # "efo"
        if self.state.efo_thresholds is None:
            self.state.efo_thresholds = (efo_bin_edges[0], efo_bin_edges[-1])

        self.efo_region = self._create_histogram_plot(
            "efo",
            self.efo_plot,
            n_efo,
            efo_bin_edges,
            efo_bin_centers,
            efo_bin_width,
            axis_range=self.efo_range,
            brush="b",
            fmt="{value:.0f}",
            support_thresholding=True,
            thresholds=self.state.efo_thresholds,
        )
        self.efo_plot.show()

        # cfr
        if self.state.cfr_thresholds is None:
            self.state.cfr_thresholds = (cfr_bin_edges[0], cfr_bin_edges[-1])

        self.cfr_region = self._create_histogram_plot(
            "cfr",
            self.cfr_plot,
            n_cfr,
            cfr_bin_edges,
            cfr_bin_centers,
            cfr_bin_width,
            axis_range=self.cfr_range,
            brush="r",
            fmt="{value:.2f}",
            support_thresholding=True,
            thresholds=self.state.cfr_thresholds,
            force_min_x_range_to_zero=False,
        )
        self.cfr_plot.show()

        # sx
        n_sx, sx_bin_edges, sx_bin_centers, sx_bin_width = prepare_histogram(
            self.processor.filtered_dataframe_stats["sx"].values,
            auto_bins=True,
            bin_size=0.0,
        )
        _ = self._create_histogram_plot(
            "sx",
            self.sx_plot,
            n_sx,
            sx_bin_edges,
            sx_bin_centers,
            sx_bin_width,
            axis_range=self.loc_precision_range,
            brush="k",
            support_thresholding=False,
        )
        self._add_median_line(
            self.sx_plot, self.processor.filtered_dataframe_stats["sx"].values
        )
        self.sx_plot.show()

        # sy
        n_sy, sy_bin_edges, sy_bin_centers, sy_bin_width = prepare_histogram(
            self.processor.filtered_dataframe_stats["sy"].values,
            auto_bins=True,
            bin_size=0.0,
        )
        _ = self._create_histogram_plot(
            "sy",
            self.sy_plot,
            n_sy,
            sy_bin_edges,
            sy_bin_centers,
            sy_bin_width,
            axis_range=self.loc_precision_range,
            brush="k",
            support_thresholding=False,
        )
        self._add_median_line(
            self.sy_plot, self.processor.filtered_dataframe_stats["sy"].values
        )
        self.sy_plot.show()

        # sz
        if self.processor.is_3d:
            n_sz, sz_bin_edges, sz_bin_centers, sz_bin_width = prepare_histogram(
                self.processor.filtered_dataframe_stats["sz"].values,
                auto_bins=True,
                bin_size=0.0,
            )
            _ = self._create_histogram_plot(
                "sz",
                self.sz_plot,
                n_sz,
                sz_bin_edges,
                sz_bin_centers,
                sz_bin_width,
                axis_range=self.loc_precision_range,
                brush="k",
                support_thresholding=False,
            )
            self._add_median_line(
                self.sz_plot,
                self.processor.filtered_dataframe_stats["sz"].values,
            )
            self.sz_plot.show()
        else:
            self.sz_plot.hide()

        # Announce that the plotting has completed
        self.plotting_completed.emit()

    def _create_histogram_plot(
        self,
        plot_id,
        plot_widget,
        n: np.ndarray,
        bin_edges: np.ndarray,
        bin_centers: np.ndarray,
        bin_width: np.ndarray,
        *,
        axis_range: Optional[tuple] = None,
        brush: str = "b",
        fmt: str = "{value:0.2f}",
        support_thresholding: bool = False,
        thresholds: Optional[Tuple] = None,
        force_min_x_range_to_zero: bool = True,
    ):
        """Create a histogram plot and return it to be added to the layout."""

        # Check for consistency
        if support_thresholding and thresholds is None:
            raise ValueError(
                "If 'support_thresholding' is True, 'thresholds' must be a tuple with two values."
            )
        if thresholds is not None and len(thresholds) != 2:
            raise ValueError(
                "If 'support_thresholding' is True, 'thresholds' must be a tuple with two values."
            )

        chart = pg.BarGraphItem(
            x=bin_centers, height=n, width=0.9 * bin_width, brush=brush
        )

        if axis_range is None:
            axis_range = (bin_edges[0], bin_edges[-1])

        # Range values
        if force_min_x_range_to_zero:
            x0 = 0.0
        else:
            x0 = axis_range[0]
        plot_widget.setXRange(x0, axis_range[1])  # setXRange()'s padding misbehaves
        plot_widget.setYRange(0.0, n.max())
        plot_widget.addItem(chart)

        region = None
        if support_thresholding:
            # Create a linear region for setting filtering thresholds
            region = pg.LinearRegionItem(
                values=[thresholds[0], thresholds[1]],
                pen={"color": "k", "width": 3},
            )

            # Mark region with data label for callbacks
            region.data_label = plot_id

            # Add to plot
            plot_widget.addItem(region)

            # Add labels with current values of lower and upper thresholds. Attach them
            # to the region to be able to access them from callbacks.
            region.low_thresh_label = pg.InfLineLabel(
                region.lines[0], fmt, position=0.95
            )
            self._change_region_label_font(region.low_thresh_label)
            region.high_thresh_label = pg.InfLineLabel(
                region.lines[1], fmt, position=0.90
            )
            self._change_region_label_font(region.high_thresh_label)

            # Connect signals
            region.sigRegionChanged.connect(self.region_pos_changed)
            region.sigRegionChangeFinished.connect(self.region_pos_changed_finished)

        # Make sure the viewbox remembers its own y range
        viewbox = plot_widget.getPlotItem().getViewBox()
        viewbox.y_min = 0.0
        viewbox.y_max = n.max()
        viewbox.sigXRangeChanged.connect(self.fix_viewbox_y_range)

        # Mark viewbox with data label for callbacks
        viewbox.data_label = plot_id

        return region

    def histogram_raise_context_menu(self, ev):
        """Create a context menu on the efo vs cfr scatter/histogram plot ROI."""
        if ev.button() == Qt.MouseButton.RightButton:
            menu = QMenu()
            reset_action = QAction("Reset default axis range")
            reset_action.triggered.connect(
                lambda checked: self.reset_default_axis_range(ev.currentItem)
            )
            menu.addAction(reset_action)
            shift_action = QAction("Move x axis origin to 0")
            shift_action.triggered.connect(
                lambda checked: self.shift_x_axis_origin_to_zero(ev.currentItem)
            )
            menu.addAction(shift_action)
            menu.addSeparator()
            export_action = QAction("Export plot")
            export_action.triggered.connect(
                lambda checked: export_plot_interactive(ev.currentItem)
            )
            menu.addAction(export_action)
            pos = ev.screenPos()
            menu.exec(QPoint(int(pos.x()), int(pos.y())))
            ev.accept()
        else:
            ev.ignore()

    def histogram_raise_context_menu_with_filtering(self, ev):
        """Create a context menu on the efo vs cfr scatter/histogram plot ROI."""
        if ev.button() == Qt.MouseButton.RightButton:
            menu = QMenu()
            ranges_action = QAction("Set ROI ranges")
            ranges_action.triggered.connect(
                lambda checked: self.roi_open_ranges_dialog(ev.currentItem.data_label)
            )
            menu.addAction(ranges_action)
            filter_action = QAction("Filter")
            filter_action.triggered.connect(
                lambda checked: self.trigger_filter_action(ev.currentItem)
            )
            menu.addAction(filter_action)
            menu.addSeparator()
            reset_action = QAction("Reset default axis range")
            reset_action.triggered.connect(
                lambda checked: self.reset_default_axis_range(ev.currentItem)
            )
            menu.addAction(reset_action)
            shift_action = QAction("Move x axis origin to 0")
            shift_action.triggered.connect(
                lambda checked: self.shift_x_axis_origin_to_zero(ev.currentItem)
            )
            menu.addAction(shift_action)
            menu.addSeparator()
            export_action = QAction("Export plot")
            export_action.triggered.connect(
                lambda checked: export_plot_interactive(ev.currentItem)
            )
            menu.addAction(export_action)
            pos = ev.screenPos()
            menu.exec(QPoint(int(pos.x()), int(pos.y())))
            ev.accept()
        else:
            ev.ignore()

    def roi_open_ranges_dialog(self, item):
        """Open dialog to manually set the filter ranges"""
        if self.roi_ranges_dialog is None:
            self.roi_ranges_dialog = ROIRanges()
            self.roi_ranges_dialog.data_ranges_changed.connect(
                self.roi_changes_finished
            )
        else:
            self.roi_ranges_dialog.update_fields()
        self.roi_ranges_dialog.set_target(item)
        self.roi_ranges_dialog.show()
        self.roi_ranges_dialog.activateWindow()

    def _add_median_line(self, plot, values):
        """Add median line to plot (with median +/- mad as label)."""
        _, _, med, mad = get_robust_threshold(values, 0.0)
        line = pg.InfiniteLine(
            pos=med,
            movable=False,
            angle=90,
            pen={"color": (200, 50, 50), "width": 3},
            label=f"median={med:.2f} ± {mad:.2f} nm",
            labelOpts={
                "position": 0.95,
                "color": (200, 50, 50),
                "fill": (200, 50, 50, 10),
                "movable": True,
            },
        )
        plot.addItem(line)

    def shift_x_axis_origin_to_zero(self, item):
        """Set the lower range of the x axis of the passed viewbox to 0."""
        if isinstance(item, ViewBox):
            view_box = item
        elif isinstance(item, AxisItem):
            view_box = item.getViewBox()
        else:
            return
        view_range = view_box.viewRange()
        view_box.setRange(xRange=(0.0, view_range[0][1]))

    def reset_default_axis_range(self, item):
        if isinstance(item, ViewBox):
            view_box = item
        elif isinstance(item, AxisItem):
            view_box = item.getViewBox()
        else:
            return
        if item.data_label == "efo":
            view_box.setRange(xRange=(self.state.efo_range[0], self.state.efo_range[1]))
            self.efo_range = self.state.efo_range
        elif item.data_label == "cfr":
            view_box.setRange(xRange=(self.state.cfr_range[0], self.state.cfr_range[1]))
            self.cfr_range = self.state.cfr_range
        elif item.data_label in ["sx", "sy", "sz"]:
            view_box.setRange(
                xRange=(
                    self.state.loc_precision_range[0],
                    self.state.loc_precision_range[1],
                )
            )
        else:
            pass

    def trigger_filter_action(self, item):
        """Set the lower range of the x axis of the passed viewbox to 0."""
        if isinstance(item, ViewBox):
            view_box = item
        elif isinstance(item, AxisItem):
            view_box = item.getViewBox()
        else:
            return
        plot_id = view_box.data_label
        if plot_id == "efo":
            self.run_efo_filter_and_broadcast_viewers_update()
        elif plot_id == "cfr":
            self.run_cfr_filter_and_broadcast_viewers_update()
        else:
            raise ValueError(f"Unexpected `plot_id` {plot_id}.")

    def region_pos_changed(self, item):
        """Called when the line region on one of the histogram plots is changing."""

        # This seems to be a bug in pyqtgraph. When moving a LinearRegionItems with
        # two InfLineLabels attached to the region boundaries (InfiniteLine), only
        # the label of the upper bound is automatically updated.
        region = item.getRegion()
        low_thresh_label = item.low_thresh_label
        if low_thresh_label.format == "{value:.2f}":
            value = f"{min(region):.2f}"
        elif low_thresh_label.format == "{value:.0f}":
            value = f"{min(region):.0f}"
        else:
            value = f"{min(region)}"
        low_thresh_label.textItem.setPlainText(value)

    @Slot(None, name="roi_changes_finished")
    def roi_changes_finished(self):
        """Called when the ROIChanges dialog has accepted the changes."""

        # Signal blocker on self.efo_plot and self.cfr_plot
        cfr_plot_blocker = QSignalBlocker(self.cfr_plot)
        efo_plot_blocker = QSignalBlocker(self.efo_plot)

        # Block signals from self.efo_plot and self.cfr_plot
        cfr_plot_blocker.reblock()
        efo_plot_blocker.reblock()

        # Update the thresholds in the EFO and CFR histograms.
        self.efo_region.setRegion(self.state.efo_thresholds)
        self.cfr_region.setRegion(self.state.cfr_thresholds)

        # Unblock the self.efo_plot and self.cfr_plot signals
        cfr_plot_blocker.unblock()
        efo_plot_blocker.unblock()

    def region_pos_changed_finished(self, item):
        """Called when the line region on one of the histogram plots has changed."""
        if item.data_label not in ["efo", "cfr"]:
            raise ValueError(f"Unexpected data label {item.data_label}.")

        # Update the correct thresholds
        if item.data_label == "efo":
            self.state.efo_thresholds = item.getRegion()
            self.efo_bounds_changed.emit()
        else:
            self.state.cfr_thresholds = item.getRegion()
            self.cfr_bounds_changed.emit()

    @staticmethod
    def _change_region_label_font(region_label):
        """Change the region label font style."""
        text_item = region_label.textItem
        text_item.setDefaultTextColor(QColor("black"))
        font = text_item.font()
        font.setWeight(QFont.Bold)
        font.setPointSize(20)

    def fix_viewbox_y_range(self, viewbox, x_range_limits):
        """Reset the y axis range whenever the x range changes."""
        viewbox.setYRange(viewbox.y_min, viewbox.y_max)
        viewbox.setAutoVisible(y=True)

    def change_efo_bounds(self):
        """Update the efo bounds on the histogram (without broadcasting any events)."""

        # Signal blocker on self.efo_plot
        efo_plot_blocker = QSignalBlocker(self.efo_plot)

        # Block signals from self.efo_plot and self.cfr_plot
        efo_plot_blocker.reblock()

        # Update the thresholds in the EFO and CFR histograms.
        self.efo_region.setRegion(self.state.efo_thresholds)

        # Unblock the self.efo_plot and self.cfr_plot signals
        efo_plot_blocker.unblock()

    def change_cfr_bounds(self):
        """Update the cfr bounds on the histogram (without broadcasting any events)."""

        # Signal blocker on self.efo_plot
        cfr_plot_blocker = QSignalBlocker(self.cfr_plot)

        # Block signals from self.efo_plot and self.cfr_plot
        cfr_plot_blocker.reblock()

        # Update the thresholds in the EFO and CFR histograms.
        self.cfr_region.setRegion(self.state.cfr_thresholds)

        # Unblock the self.efo_plot and self.cfr_plot signals
        cfr_plot_blocker.unblock()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from PySide6.QtCore import Signal, Slot
from PySide6.QtGui import QDoubleValidator, QIntValidator
from PySide6.QtWidgets import QDialog, QHBoxLayout, QMessageBox, QTextEdit

from pyminflux import __APP_NAME__
from pyminflux.state import State

from ..settings import Settings
from .ui_options import Ui_Options


class Options(QDialog, Ui_Options):
    # Signal that the options have changed
    min_trace_length_option_changed = Signal(name="min_trace_length_option_changed")
    efo_bin_size_hz_option_changed = Signal(name="efo_bin_size_hz_option_changed")
    weigh_avg_localization_by_eco_option_changed = Signal(
        name="weigh_avg_localization_by_eco_option_changed"
    )

    def __init__(self, parent=None):
        """Constructor."""

        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_Options()
        self.ui.setupUi(self)

        # Keep a reference to the singleton State class
        self.state = State()

        # Add the help text edit
        self.ui.teHelp.setReadOnly(True)
        font_metrics = self.ui.teHelp.fontMetrics()
        self.ui.teHelp.setMaximumHeight(font_metrics.height() * 3)
        self.ui.teHelp.setText('Click on a "?" button to get help.')

        # Hide the border and change the background color
        self.ui.teHelp.setStyleSheet(
            """
            QTextEdit {
                border: none;
                background-color: transparent;
                font-style: italic;
            }
        """
        )

        # Keep track of the validity of all entries
        self.valid = {
            "leMinTIDNum": True,
            "leEFOBinSize": True,
            "leEFOSingleEmitterFrequency": True,
            "cbWeightAvgLocByECO": True,
            "leEFORangeMin": True,
            "leEFORangeMax": True,
            "leCFRRangeMin": True,
            "leCFRRangeMax": True,
            "leLocPrecRangeMin": True,
            "leLocPrecRangeMax": True,
            "leZScalingFactor": True,
            "lePlotExportDPI": True,
        }

        # Set defaults
        self.ui.leMinTIDNum.setText(str(self.state.min_trace_length))
        self.ui.leMinTIDNum.setValidator(QIntValidator(bottom=0))
        self.ui.leEFOBinSize.setText(str(self.state.efo_bin_size_hz))
        self.ui.leEFOBinSize.setValidator(QDoubleValidator(bottom=0.0))
        self.ui.leEFOSingleEmitterFrequency.setText(
            str(self.state.efo_expected_frequency)
        )
        self.ui.leEFOSingleEmitterFrequency.setValidator(QDoubleValidator(bottom=0.0))
        self.ui.cbWeightAvgLocByECO.setChecked(self.state.weigh_avg_localization_by_eco)

        if self.state.efo_range is None:
            self.ui.leEFORangeMin.setText("")
            self.ui.leEFORangeMax.setText("")
        else:
            self.ui.leEFORangeMin.setText(str(self.state.efo_range[0]))
            self.ui.leEFORangeMax.setText(str(self.state.efo_range[1]))
        self.ui.leEFORangeMin.setValidator(QDoubleValidator(bottom=0.0, decimals=1))
        self.ui.leEFORangeMax.setValidator(QDoubleValidator(bottom=0.0, decimals=1))

        if self.state.cfr_range is None:
            self.ui.leCFRRangeMin.setText("")
            self.ui.leCFRRangeMax.setText("")
        else:
            self.ui.leCFRRangeMin.setText(str(self.state.cfr_range[0]))
            self.ui.leCFRRangeMax.setText(str(self.state.cfr_range[1]))
        self.ui.leCFRRangeMin.setValidator(QDoubleValidator(bottom=0.0, decimals=1))
        self.ui.leCFRRangeMax.setValidator(QDoubleValidator(bottom=0.0, decimals=1))

        if self.state.loc_precision_range is None:
            self.ui.leLocPrecRangeMin.setText("")
            self.ui.leLocPrecRangeMax.setText("")
        else:
            self.ui.leLocPrecRangeMin.setText(str(self.state.loc_precision_range[0]))
            self.ui.leLocPrecRangeMax.setText(str(self.state.loc_precision_range[1]))
        self.ui.leLocPrecRangeMin.setValidator(QDoubleValidator(bottom=0.0, decimals=1))
        self.ui.leLocPrecRangeMax.setValidator(QDoubleValidator(bottom=0.0, decimals=1))

        self.ui.leZScalingFactor.setText(str(self.state.z_scaling_factor))
        self.ui.leZScalingFactor.setValidator(QDoubleValidator(bottom=0.0, decimals=2))

        self.ui.lePlotExportDPI.setText(str(self.state.plot_export_dpi))
        self.ui.lePlotExportDPI.setValidator(QIntValidator(bottom=72))

        self.ui.cbOpenConsoleAtStart.setChecked(self.state.open_console_at_start)

        # Set signal-slot connections
        self.setup_conn()

    def setup_conn(self):
        """Set up signal-slot connections."""
        self.ui.leMinTIDNum.textChanged.connect(self.persist_min_trace_length)
        self.ui.pbSetDefault.clicked.connect(self.set_as_new_default)
        self.ui.cbWeightAvgLocByECO.stateChanged.connect(
            self.persist_weigh_avg_localization_by_eco
        )
        self.ui.leEFOBinSize.textChanged.connect(self.persist_efo_bin_size_hz)
        self.ui.leEFOSingleEmitterFrequency.textChanged.connect(
            self.persist_efo_expected_cutoff
        )

        self.ui.leCFRRangeMin.textChanged.connect(self.persist_cfr_range)
        self.ui.leCFRRangeMax.textChanged.connect(self.persist_cfr_range)
        self.ui.leEFORangeMin.textChanged.connect(self.persist_efo_range)
        self.ui.leEFORangeMax.textChanged.connect(self.persist_efo_range)
        self.ui.leLocPrecRangeMin.textChanged.connect(self.persist_loc_prec_range)
        self.ui.leLocPrecRangeMax.textChanged.connect(self.persist_loc_prec_range)

        self.ui.leZScalingFactor.textChanged.connect(self.persist_z_scaling_factor)

        self.ui.lePlotExportDPI.textChanged.connect(self.persist_plot_export_dpi)

        self.ui.cbOpenConsoleAtStart.stateChanged.connect(
            self.persist_open_console_at_start
        )

        self.ui.pbMinTIDNumHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Minimum number of localizations within a trace to be considered for "
                "plotting and precision calculation."
            )
        )
        self.ui.pbZScalingFactorHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Scales the z positions to compensate for the refractive index mismatch "
                "between the coverglass and the sample."
            )
        )
        self.ui.pbEFOBinSizeHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Bin size to be used for EFO distribution plotting on the Analyzer. "
                "Set to 0 for automatic estimation."
            )
        )
        self.ui.pbEFOSingleEmitterFrequencyHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Expected default frequency for single emitters to be used for automated "
                "EFO threshold selection."
            )
        )
        self.ui.pbEFORangeHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Default plot range for the EFO histogram in the Analyzer."
            )
        )
        self.ui.pbCFRRangeHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Default plot range for the CFR histogram in the Analyzer."
            )
        )
        self.ui.pbLocPrecRangeHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Default plot range for the localization precision histograms in the Analyzer."
            )
        )
        self.ui.pbWeightAvgLocByECOHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Whether to use the ECO value as weighting factor during the calculation of the "
                "average localization position for each trace ID."
            )
        )
        self.ui.pbPlotExportDPIHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Resolution (in dots per inch) for the all plots when exported as .png images."
            )
        )
        self.ui.pbOpenConsoleAtStartHelp.clicked.connect(
            lambda _: self.ui.teHelp.setText(
                "Whether to open the console at application start."
            )
        )

    @Slot(str, name="persist_min_trace_length")
    def persist_min_trace_length(self, text):
        try:
            min_trace_length = int(text)
        except Exception as _:
            self.ui.leMinTIDNum.setStyleSheet("background-color: red;")
            self.valid["leMinTIDNum"] = False
            return
        self.ui.leMinTIDNum.setStyleSheet("")
        self.valid["leMinTIDNum"] = True
        self.state.min_trace_length = min_trace_length

        # Signal the change
        self.min_trace_length_option_changed.emit()

    @Slot(str, name="persist_z_scaling_factor")
    def persist_z_scaling_factor(self, text):
        try:
            z_scaling_factor = float(text)
        except Exception as _:
            self.ui.leZScalingFactor.setStyleSheet("background-color: red;")
            self.valid["leZScalingFactor"] = False
            return
        self.ui.leZScalingFactor.setStyleSheet("")
        self.valid["leZScalingFactor"] = True
        self.state.z_scaling_factor = z_scaling_factor

    @Slot(str, name="persist_plot_export_dpi")
    def persist_plot_export_dpi(self, text):
        try:
            plot_export_dpi = int(text)
        except Exception as _:
            self.ui.lePlotExportDPI.setStyleSheet("background-color: red;")
            self.valid["lePlotExportDPI"] = False
            return
        self.ui.lePlotExportDPI.setStyleSheet("")
        self.valid["lePlotExportDPI"] = True
        self.state.plot_export_dpi = plot_export_dpi

    @Slot(str, name="persist_efo_bin_size_hz")
    def persist_efo_bin_size_hz(self, text):
        try:
            efo_bin_size_hz = float(text)
        except Exception as _:
            self.ui.leEFOBinSize.setStyleSheet("background-color: red;")
            self.valid["hlEFOBinSize"] = False
            return
        self.ui.leEFOBinSize.setStyleSheet("")
        self.valid["hlEFOBinSize"] = True
        self.state.efo_bin_size_hz = efo_bin_size_hz

        # Signal the change
        self.efo_bin_size_hz_option_changed.emit()

    @Slot(str, name="persist_efo_expected_cutoff")
    def persist_efo_expected_cutoff(self, text):
        try:
            efo_expected_frequency = float(text)
        except Exception as _:
            self.ui.leEFOSingleEmitterFrequency.setStyleSheet("background-color: red;")
            self.valid["leEFOSingleEmitterFrequency"] = False
            return
        self.ui.leEFOSingleEmitterFrequency.setStyleSheet("")
        self.valid["leEFOSingleEmitterFrequency"] = True
        self.state.efo_expected_frequency = efo_expected_frequency

    @Slot(str, name="persist_cfr_range")
    def persist_cfr_range(self, _):
        """Persist CFR range."""

        # Initialize the status to valid
        self.valid["leCFRRangeMin"] = True
        self.valid["leCFRRangeMax"] = True

        # Get current values
        cfr_min = self.ui.leCFRRangeMin.text()
        cfr_max = self.ui.leCFRRangeMax.text()

        # Validate them
        cfr_min = self._validate(cfr_min, self.ui.leCFRRangeMin)
        cfr_max = self._validate(cfr_max, self.ui.leCFRRangeMax)

        # Check that min and max values are consistent
        if cfr_min != "" and cfr_max != "":
            if cfr_min < cfr_max:
                self.state.cfr_range = (cfr_min, cfr_max)
                stylesheet = ""
            else:
                stylesheet = "background-color: red;"
                self.valid["leCFRRangeMin"] = False
                self.valid["leCFRRangeMax"] = False
                self.state.cfr_range = None
            self.ui.leCFRRangeMin.setStyleSheet(stylesheet)
            self.ui.leCFRRangeMax.setStyleSheet(stylesheet)
        elif cfr_min == "" and cfr_max == "":
            self.state.cfr_range = None
        else:
            self.state.cfr_range = None
            if cfr_min == "":
                self.ui.leCFRRangeMin.setStyleSheet("background-color: red;")
                self.valid["leCFRRangeMin"] = False
            if cfr_max == "":
                self.ui.leCFRRangeMax.setStyleSheet("background-color: red;")
                self.valid["leCFRRangeMax"] = False

    @Slot(str, name="persist_efo_range")
    def persist_efo_range(self, _):
        """Persist EFO range."""

        # Initialize the status to valid
        self.valid["leEFORangeMin"] = True
        self.valid["leEFORangeMax"] = True

        # Get current values
        efo_min = self.ui.leEFORangeMin.text()
        efo_max = self.ui.leEFORangeMax.text()

        # Validate them
        efo_min = self._validate(efo_min, self.ui.leEFORangeMin)
        efo_max = self._validate(efo_max, self.ui.leEFORangeMax)

        # Check that min and max values are consistent
        if efo_min != "" and efo_max != "":
            if efo_min < efo_max:
                self.state.efo_range = (efo_min, efo_max)
                stylesheet = ""
            else:
                stylesheet = "background-color: red;"
                self.valid["leEFORangeMin"] = False
                self.valid["leEFORangeMax"] = False
                self.state.efo_range = None
            self.ui.leEFORangeMin.setStyleSheet(stylesheet)
            self.ui.leEFORangeMax.setStyleSheet(stylesheet)
        elif efo_min == "" and efo_max == "":
            self.state.efo_range = None
        else:
            self.state.efo_range = None
            if efo_min == "":
                self.ui.leEFORangeMin.setStyleSheet("background-color: red;")
                self.valid["leEFORangeMin"] = False
            if efo_max == "":
                self.ui.leEFORangeMax.setStyleSheet("background-color: red;")
                self.valid["leEFORangeMax"] = False

    @Slot(str, name="persist_loc_prec_range")
    def persist_loc_prec_range(self, _):
        """Persist localization precision range."""

        # Initialize the status to valid
        self.valid["leLocPrecRangeMin"] = True
        self.valid["leLocPrecRangeMax"] = True

        # Get current values
        loc_prec_min = self.ui.leLocPrecRangeMin.text()
        loc_prec_max = self.ui.leLocPrecRangeMax.text()

        # Validate them
        loc_prec_min = self._validate(loc_prec_min, self.ui.leLocPrecRangeMin)
        loc_prec_max = self._validate(loc_prec_max, self.ui.leLocPrecRangeMax)

        # Check that min and max values are consistent
        if loc_prec_min != "" and loc_prec_max != "":
            if loc_prec_min < loc_prec_max:
                self.state.loc_precision_range = (loc_prec_min, loc_prec_max)
                stylesheet = ""
            else:
                stylesheet = "background-color: red;"
                self.valid["leLocPrecRangeMin"] = False
                self.valid["leLocPrecRangeMax"] = False
                self.state.loc_precision_range = None
            self.ui.leLocPrecRangeMin.setStyleSheet(stylesheet)
            self.ui.leLocPrecRangeMax.setStyleSheet(stylesheet)
        elif loc_prec_min == "" and loc_prec_max == "":
            self.state.loc_precision_range = None
        else:
            self.state.loc_precision_range = None
            if loc_prec_min == "":
                self.ui.leLocPrecRangeMin.setStyleSheet("background-color: red;")
                self.valid["leLocPrecRangeMin"] = False
            if loc_prec_max == "":
                self.ui.leLocPrecRangeMax.setStyleSheet("background-color: red;")
                self.valid["leLocPrecRangeMax"] = False

    @Slot(str, name="persist_weigh_avg_localization_by_eco")
    def persist_weigh_avg_localization_by_eco(self, state):
        self.state.weigh_avg_localization_by_eco = state != 0

        # Signal the change
        self.weigh_avg_localization_by_eco_option_changed.emit()

    @Slot(str, name="persist_open_console_at_start")
    def persist_open_console_at_start(self, state):
        self.state.open_console_at_start = state != 0

    @Slot(str, name="set_as_new_default")
    def set_as_new_default(self, text):
        """Persist current selection as new default options."""

        # Only save if all fields are properly set
        all_true = all(self.valid.values())

        if not all_true:
            error_dialog = QMessageBox()
            error_dialog.setIcon(QMessageBox.Critical)
            error_dialog.setWindowTitle("Error")
            error_dialog.setText("Please fix all invalid entries before saving!")
            error_dialog.setStandardButtons(QMessageBox.Ok)
            error_dialog.exec_()
            return

        # Update the application settings
        settings = Settings()
        settings.instance.setValue(
            "options/min_trace_length", int(self.state.min_trace_length)
        )
        settings.instance.setValue(
            "options/z_scaling_factor", float(self.state.z_scaling_factor)
        )
        settings.instance.setValue(
            "options/efo_bin_size_hz", float(self.state.efo_bin_size_hz)
        )
        settings.instance.setValue(
            "options/efo_expected_frequency", float(self.state.efo_expected_frequency)
        )
        settings.instance.setValue(
            "options/weigh_avg_localization_by_eco",
            self.state.weigh_avg_localization_by_eco,
        )
        if self.state.efo_range is None:
            settings.instance.setValue("options/efo_range", "None")
        else:
            settings.instance.setValue("options/efo_range", list(self.state.efo_range))
        if self.state.cfr_range is None:
            settings.instance.setValue("options/cfr_range", "None")
        else:
            settings.instance.setValue("options/cfr_range", list(self.state.cfr_range))
        if self.state.loc_precision_range is None:
            settings.instance.setValue("options/loc_precision_range", "None")
        else:
            settings.instance.setValue(
                "options/loc_precision_range", list(self.state.loc_precision_range)
            )
        settings.instance.setValue(
            "options/open_console_at_start",
            self.state.open_console_at_start,
        )
        settings.instance.setValue(
            "options/plot_export_dpi", int(self.state.plot_export_dpi)
        )

        # This property was superseded by "options/min_trace_length"
        if settings.instance.contains("options/min_num_loc_per_trace"):
            settings.instance.remove("options/min_num_loc_per_trace")

        # Sync the settings with file
        settings.instance.sync()

    def _validate(self, value, line_edit):
        """Check that the value in the QLineEdit is a valid float, or reset it and visually mark it otherwise."""
        if value == "":
            line_edit.setStyleSheet("")
            return value

        try:
            value = float(value)
            line_edit.setStyleSheet("")
        except ValueError:
            value = ""
            line_edit.setText("")
            line_edit.setStyleSheet("background-color: red;")
        return value
